import logging
import os
import json
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from fitxf.math.algo.encoding.Base64 import Base64
from fitxf.math.fit.utils.FitUtils import FitUtils
from sklearn.decomposition import PCA
from fitxf.math.fit.utils.TensorUtils import TensorUtils
from fitxf.math.utils.Env import Env
from fitxf.math.utils.Logging import Logging


#
# A general interface to transform original points X with associated labels into a transform X_ = f(X)
# X_ is possibly a compressed version, but not necessarily.
# If X_ is a clustered form, this means points in the same cluster will have the exact same values,
# in the usual case the center of the cluster value.
# The entire purpose of Fit transformation is mainly to compress data, to economize RAM.
# Examples:
#    Given
#       X = [[1,2,3], [3,2,1], [-1,-2,-2], [-3,-4,-2]]
#       X_labels = ['+', '+', '-', '-']
#    and arbitrary X_any = [[9,9,8], [-55,-33,-55]]
#    PCA Transform 1-dim:
#       [[ 3.72936414], [ 3.81640291], [-2.55648061], [-4.98928643]]
#       predict X_any: [['+', '+', '-', '-'], ['-', '-', '+', '+']]
#    Cluster Transform:
#
class FitXformInterface:

    KEY_X_TRANSFORM = 'X_transform'
    KEY_X_LABELS = 'X_labels'
    KEY_X_FULL_RECS = 'X_full_records'
    KEY_CENTROID = 'centroid'
    KEY_N_ITERATIONS = 'n_iter'
    KEY_N_COMPONENTS_OR_CENTERS = 'n'
    KEY_PRINCIPAL_COMPONENTS = 'principal_components'
    KEY_CENTERS = 'centers'
    KEY_GRID_VECTORS = 'grid_vectors'
    KEY_GRID_NUMBERS = 'grid_numbers'

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        self.tensor_utils = TensorUtils(logger=self.logger)
        self.base64 = Base64(logger=self.logger)

        # Source Data/labels
        self.X = None
        self.X_labels = None
        self.X_full_records = None

        # Derived transforms
        self.X_transform = None
        # Create an artificial grid
        self.X_grid_vectors = None
        self.X_grid_numbers = None
        # Measures for us to optimize the number of optimal number of pca components
        self.grid_density = None
        self.grid_density_mean = None
        # Other model dependent values
        self.model_train_total_iterations = None
        self.model_n_components_or_centers = None
        self.model_principal_components = None
        self.model_centers = None
        self.model_centroid = None
        # same as cluter inertia or if not cluster, is error of distances between point and transform summed up
        self.distance_error = None
        self.distance_error_mean = None
        self.angle_error = None
        self.angle_error_mean = None
        return

    def is_model_ready(self):
        raise Exception('Must be implemented by derived class')

    # Will dynamically look for optimal model parameters, on the condition of the target grid density
    def fit_optimal(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            # may not apply to all models. for clusters, grid is just the cluster.
            # for PCA, we need to predefine how to calculate the grid.
            target_grid_density = 2, # at least 2 points per grid
            # allowed values 'median', 'mean', 'min'
            measure = 'median',
            # Model dependent interpretation, or ignore if not relevant for specific model
            min_components = 1,
            max_components = 99999,
            # pass through mode
            test_mode = False,
    ) -> dict:
        raise Exception('Must be implemented by derived class')

    def fit(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            # Model dependent interpretation, or ignore if not relevant for specific model
            # For example, can mean how many clusters, or how many PCA components, or how many to sample
            # in a discrete Fourier transform, etc.
            n_components = 2,
            # for fine-tuning already trained clusters, thus no need to start all over again
            # useful for clusters of more than 1,000,000 points for example, where starting
            # again means another half day of fit training
            start_centers: np.ndarray = None,
            # pass through mode
            test_mode = False,
    ) -> dict:
        raise Exception('Must be implemented by derived class')

    def fine_tune(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            n_components: int = None,
            # pass through mode
            test_mode = False,
    ) -> dict:
        raise Exception('Must be implemented by derived class')

    def transform(
            self,
            X: np.ndarray,
    ) -> np.ndarray:
        raise Exception('Must be implemented by derived class')

    def inverse_transform(
            self,
            X: np.ndarray,
    ) -> np.ndarray:
        raise Exception('Must be implemented by derived class')

    def model_to_b64json(
            self,
            numpy_to_base64_str = False,
            dump_to_b64json_str = False,
    ):
        if numpy_to_base64_str:
            x_tf = self.base64.encode_numpy_array_to_base64_string_multidim(
                x = self.X_transform,
                data_type = np.float64,
            )
            centroid = self.base64.encode_numpy_array_to_base64_string_multidim(
                x = self.model_centroid,
                data_type = np.float64,
            )
            principle_components = self.base64.encode_numpy_array_to_base64_string_multidim(
                x = self.model_principal_components,
                data_type = np.float64,
            )
            centers = self.base64.encode_numpy_array_to_base64_string_multidim(
                x = self.model_centers,
                data_type = np.float64,
            )
            grid_vecs = self.base64.encode_numpy_array_to_base64_string_multidim(
                x = self.X_grid_vectors,
                data_type = np.int64,
            )
            grid_numbers = self.base64.encode_numpy_array_to_base64_string_multidim(
                x = self.X_grid_numbers,
                data_type = np.int64,
            )
        else:
            x_tf = self.X_transform
            centroid = self.model_centroid
            principle_components = self.model_principal_components
            centers = self.model_centers
            grid_vecs = self.X_grid_vectors
            grid_numbers = self.X_grid_numbers

        model_json = {
            self.KEY_X_TRANSFORM: x_tf,
            self.KEY_X_LABELS: self.X_labels,
            self.KEY_CENTROID: centroid,
            self.KEY_N_ITERATIONS: self.model_train_total_iterations,
            self.KEY_N_COMPONENTS_OR_CENTERS: self.model_n_components_or_centers,
            self.KEY_PRINCIPAL_COMPONENTS: principle_components,
            self.KEY_CENTERS: centers,
            self.KEY_GRID_VECTORS: grid_vecs,
            self.KEY_GRID_NUMBERS: grid_numbers,
        }
        if dump_to_b64json_str:
            return self.base64.encode(
                b = str(json.dumps(model_json)).encode(encoding='utf-8'),
            )
        else:
            return model_json

    def load_model_from_b64json(
            self,
            model_b64json,
    ) -> dict:
        self.logger.debug('Loading model b64: ' + str(model_b64json))
        model_dict_json = self.base64.decode(s=model_b64json)
        model_dict = json.loads(model_dict_json)

        self.logger.debug('Loading model b64 json: ' + str(model_dict))
        self.X_transform = self.base64.decode_base64_string_to_numpy_array_multidim(
            string = model_dict[self.KEY_X_TRANSFORM],
            data_type = np.float64,
        )
        self.logger.debug('Loaded X_transform as: ' + str(self.X_transform))

        self.X_labels = model_dict[self.KEY_X_LABELS]
        self.model_centroid = self.base64.decode_base64_string_to_numpy_array_multidim(
            string = model_dict[self.KEY_CENTROID],
            data_type = np.float64,
        )
        self.logger.debug('Loaded X_labels as: ' + str(self.X_labels))

        self.model_n_components_or_centers = model_dict[self.KEY_N_COMPONENTS_OR_CENTERS]
        self.model_principal_components = self.base64.decode_base64_string_to_numpy_array_multidim(
            string = model_dict[self.KEY_PRINCIPAL_COMPONENTS],
            data_type = np.float64,
        )
        self.logger.debug('Loaded model_principal_components as: ' + str(self.model_principal_components))

        self.model_centers = self.base64.decode_base64_string_to_numpy_array_multidim(
            string = model_dict[self.KEY_CENTERS],
            data_type = np.float64,
        )
        self.logger.debug('Loaded model_centers as: ' + str(self.model_centers))

        self.X_grid_vectors = self.base64.decode_base64_string_to_numpy_array_multidim(
            string = model_dict[self.KEY_GRID_VECTORS],
            data_type = np.int64,
        )
        self.logger.debug('Loaded X_grid_vectors as: ' + str(self.X_grid_vectors))

        self.X_grid_numbers = self.base64.decode_base64_string_to_numpy_array_multidim(
            string = model_dict[self.KEY_GRID_NUMBERS],
            data_type = np.int64,
        )
        self.logger.debug('Loaded X_grid_numbers as: ' + str(self.X_grid_numbers))

        return model_dict

    def predict(
            self,
            X: np.ndarray,
            # can be cluster numbers to zoom into
            X_search_local_space: np.ndarray = None,
            labels_search_local_space: list = None,
            top_k = 5,
            return_full_record = False,
            use_grid = False,
    ):
        raise Exception('Must be implemented by derived class')

    def predict_standard(
            self,
            X: np.ndarray,
            ref_X: np.ndarray,
            ref_labels: list,
            ref_full_records: list = (),
            top_k = 5,
            similarity_type = 'cosine',
            return_full_record = False,
    ):
        """
        result_ordered returns 2D matrix, each row represents the closest indexes of ref ordered
        So means if the row is [5, 0, 1, 3, 2, 4], means that the closest match is index 5 in ref,
        2nd closest is index 0 in ref, and so on.
        m_dot_ordered is the result of the dot product, a value of magnitude equal or less than 1.
        """
        self.logger.info('For prediction, using similarity type "' + str(similarity_type) + '"')
        if similarity_type == 'distance':
            result_ordered, m_dot_ordered = self.tensor_utils.similarity_distance(
                x = X,
                ref = ref_X,
                return_tensors = 'np',
            )
        else:
            result_ordered, m_dot_ordered = self.tensor_utils.dot_sim(
                x = X,
                ref = ref_X,
                return_tensors = 'np',
            )

        self.logger.debug(
            'Result ordered for top k ' + str(top_k) + ': ' + str(result_ordered) + ', ref length ' + str(len(ref_X))
        )

        if return_full_record:
            assert ref_full_records is not None, 'Cannot return full records'
            pred_records = [[ref_full_records[i] for i in np_row] for np_row in result_ordered]
            pred_probs_list = m_dot_ordered.tolist()

            return [ar[0:min(top_k, len(ar))] for ar in pred_records], \
                [ar[0:min(top_k, len(ar))] for ar in pred_probs_list]
        else:
            pred_labels = np.array(ref_labels)[result_ordered].tolist()
            pred_probs_list = m_dot_ordered.tolist()

            return [ar[0:min(top_k, len(ar))] for ar in pred_labels], \
                [ar[0:min(top_k, len(ar))] for ar in pred_probs_list]

    def create_scatter_plot2d(
            self,
            x_transform: np.ndarray,
            labels_list,
            show = False,
            save_filepath = None,
            add_noise = False,
    ):
        if x_transform.shape[-1] != 2:
            self.logger.info('X not 2D, with shape ' + str(x_transform.shape) + '. Using PCA to reduce to 2D..')
            pca = PCA(n_components=2)
            X_2d = pca.fit_transform(pd.DataFrame(x_transform))
        else:
            X_2d = np.array(x_transform)

        if add_noise:
            fu = FitUtils(logger=self.logger)
            ds = fu.get_point_distances(np_tensors=X_2d)
            dist_median = np.median(ds)
            # 1/5 of average point distances on average
            X_2d += np.random.random(size=X_2d.shape) * dist_median / 2.5

        columns = ['PC'+str(i+1) for i in range(2)]
        # Create a new DataFrame with reduced dimensions
        df_reduced = pd.DataFrame(X_2d, columns=columns)

        plt.scatter(df_reduced['PC1'], df_reduced['PC2'], label=labels_list)

        # Add labels and title
        plt.xlabel('X')
        plt.ylabel('Y')
        plt.title('Scatter Plot')

        # Add labels to each dot
        # for i, label in enumerate(embeddings_df.iloc[:, -1].to_list()):
        for i, label in enumerate(labels_list):
            plt.text(df_reduced['PC1'][i], df_reduced['PC2'][i], label)

        if show:
            plt.show()

        # Save and display the plot
        if save_filepath is not None:
            plt.savefig(save_filepath, format='png')
        return plt


class SampleFit(FitXformInterface):

    def __init__(
            self,
            logger = None,
    ):
        super().__init__(logger=logger)
        self.model_centroid = None
        self.pca_components_2d = None
        return

    def fit(
            self,
            X: np.ndarray,
            X_labels = None,
            X_full_records = None,
            # Model dependent interpretation, or ignore if not relevant for specific model
            # For example, can mean how many clusters, or how many PCA components, or how many to sample
            # in a discrete Fourier transform, etc.
            n_components = 2,
            # for fine-tuning already trained clusters, thus no need to start all over again
            # useful for clusters of more than 1,000,000 points for example, where starting
            # again means another half day of fit training
            start_centers: np.ndarray = None,
            return_details = False,
            # pass through mode
            test_mode = False,
    ) -> dict:
        pca = PCA(n_components=n_components)

        # Create copy, don't use same ref from input X
        self.X = np.array(X)
        self.X_transform = pca.fit_transform(pd.DataFrame(self.X))
        self.X_labels = np.array(X_labels)
        self.X_full_records = X_full_records
        self.principal_components = pca.components_
        self.centroid = np.mean(X, axis=0)
        return {
            self.KEY_X_TRANSFORM: self.X_transform,
            self.KEY_PRINCIPAL_COMPONENTS: self.principal_components,
            self.KEY_N_COMPONENTS_OR_CENTERS: n_components,
        }


class FitXformInterfaceUnitTest:

    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self, plot_chart=False):
        from fitxf.math.lang.encode.LangModelPtSingleton import LangModelPtSingleton, LangModelInterface
        from fitxf.math.lang.encode.LangModelPt import LangModelPt
        ft = SampleFit(logger=self.logger)

        texts = [
            "Let's have coffee", "Free for a drink?", "How about Starbucks?",
            "I am busy", "Go away", "Don't disturb me",
            "Monetary policies", "Interest rates", "Deposit rates",
        ]
        labels = ['d', 'd', 'd', 'x', 'x', 'x', '$', '$', '$']
        lmo = LangModelPtSingleton.get_singleton(
            LmClass = LangModelPt,
            cache_folder = Env().MODELS_PRETRAINED_DIR,
            logger = self.logger,
        )

        embeddings = lmo.encode(content_list=texts, return_tensors='np')

        res = ft.fit(X=embeddings)
        x_compressed = res[ft.KEY_X_TRANSFORM]

        small_random_values = (np.random.random(size=x_compressed.shape) - 0.5) / 100
        pred_labels, pred_probs = ft.predict_standard(
            X = x_compressed + small_random_values,
            ref_X = x_compressed,
            ref_labels = labels,
        )
        prd_labels_top = [r[0] for r in pred_labels]
        assert prd_labels_top == labels, 'Predicted top labels not expected ' + str(zip(prd_labels_top, labels))
        prd_probs_top = [r[0] for r in pred_probs]
        assert np.min(np.array(prd_probs_top)) > 0.999, 'Top probabilities ' + str(prd_probs_top)

        if plot_chart:
            ft.create_scatter_plot2d(
                x_transform = embeddings,
                labels_list = zip(labels, texts),
                show = True,
            )
        print('ALL TESTS PASSED OK')
        return


if __name__ == '__main__':
    FitXformInterfaceUnitTest(logger=Logging.get_default_logger(log_level=logging.INFO, propagate=False)).test(
        plot_chart = True,
    )
    exit(0)



import logging
import pandas as pd
import numpy as np
from fitxf.math.graph.GraphUtils import GraphUtils
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Pandas import Pandas


class GraphUtilsUnitTest:

    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self):
        gu = GraphUtils(logger=self.logger)
        MAX_DIST = 999999
        edge_data = [
            {'key': 'plane', 'u': 'Shanghai', 'v': 'Tokyo', 'distance': 10, 'comment': 'Shanghai-Tokyo flight'},
            # duplicate (will not be added), order does not matter
            {'key': 'plane', 'u': 'Tokyo', 'v': 'Shanghai', 'distance': 22, 'comment': 'Tokyo-Shanghai flight'},
            # teleport path Tokyo --> Beijing --> Shanghai shorter distance than plane Tokyo --> Shanghai
            {'key': 'teleport', 'u': 'Tokyo', 'v': 'Beijing', 'distance': 2, 'comment': 'Tokyo-Beijing teleport'},
            {'key': 'plane', 'u': 'Tokyo', 'v': 'Beijing', 'distance': 9, 'comment': 'Tokyo-Beijing plane'},
            {'key': 'teleport', 'u': 'Beijing', 'v': 'Shanghai', 'distance': 1, 'comment': 'Beijing-Shanghai teleport'},
            # Other paths
            {'key': 'ship', 'u': 'Shanghai', 'v': 'Tokyo', 'distance': 100, 'comment': 'Shanghai-Tokyo sea'},
            {'key': 'plane', 'u': 'Moscow', 'v': 'Tokyo', 'distance': 100, 'comment': 'Asia-Russia flight'},
            {'key': 'train', 'u': 'Moscow', 'v': 'Tokyo', 'distance': 10000, 'comment': 'Asia-Russia train'},
            {'key': 'ship', 'u': 'Moscow', 'v': 'Tokyo', 'distance': MAX_DIST, 'comment': 'Asia-Russia sea'},
            {'key': 'plane', 'u': 'Medellin', 'v': 'Antartica', 'distance': 888, 'comment': 'Medellin-Antartica'},
        ]
        G_test = {}
        for directed, exp_total_edges, exp_nodes in [
            (False, 9, ['Antartica', 'Beijing', 'Medellin', 'Moscow', 'Shanghai', 'Tokyo']),
            (True, 10, ['Antartica', 'Beijing', 'Medellin', 'Moscow', 'Shanghai', 'Tokyo']),
        ]:
            G_tmp = gu.create_multi_graph(
                edges = edge_data,
                col_weight = 'distance',
                directed = directed,
            )
            G_test[directed] = G_tmp
            print('-----------------------------------------------------------------------------')
            print('Edges (directed=' + str(directed) + ')')
            [print(i, G_tmp.get_edge_data(u=u, v=v)) for i, (u, v, key) in enumerate(G_tmp.edges)]
            all_edges = list(G_tmp.edges)
            all_edges.sort()
            assert len(G_tmp.edges) == exp_total_edges, \
                'Directed ' + str(directed) + ' Expect ' + str(exp_total_edges) + ' edges, but got ' \
                + str(len(G_tmp.edges))
            print('-----------------------------------------------------------------------------')
            print('Nodes (directed=' + str(directed) + ')')
            print(G_tmp.nodes)
            all_nodes = list(G_tmp.nodes)
            all_nodes.sort()
            assert all_nodes == ['Antartica', 'Beijing', 'Medellin', 'Moscow', 'Shanghai', 'Tokyo'], \
                'Directed ' + str(directed) + ' Nodes not expected ' + str(all_nodes)

        paths_dijkstra = {}
        paths_shortest = {}
        for dir in [True, False]:
            paths_dijkstra[dir] = gu.get_dijkstra_path_all_pairs(G=G_test[dir])
            print('-----------------------------------------------------------------------------')
            print('Dijkstra Paths (directed = ' + str(dir) + ')')
            print(pd.DataFrame.from_records([{'edge': k, 'dijkstra-path': v} for k, v in paths_dijkstra[dir].items()]))

            paths_shortest[dir] = gu.get_shortest_path_all_pairs(G=G_test[dir])
            print('-----------------------------------------------------------------------------')
            print('Shortest Paths (directed = ' + str(dir) + ')')
            print(pd.DataFrame.from_records([{'edge': k, 'shortest-path': v} for k, v in paths_shortest[dir].items()]))

        for dir, edge, exp_best_path in [
            # teleport path for undirected graph from Shanghai-->Beijing-->Tokyo is fastest
            (False, ('Shanghai', 'Tokyo'), ['Shanghai', 'Beijing', 'Tokyo']),
            (False, ('Shanghai', 'Moscow'), ['Shanghai', 'Beijing', 'Tokyo', 'Moscow']),
            # no teleport path for directed graph from Shanghai-->Tokyo
            (True, ('Shanghai', 'Tokyo'), ['Shanghai', 'Tokyo']),
            (True, ('Shanghai', 'Moscow'), None),
        ]:
            observed_edge = paths_dijkstra[dir].get(edge, None)
            assert observed_edge == exp_best_path, \
                'Directed "' + str(dir) + '" Edge ' + str(edge) + ' path ' + str(observed_edge) \
                + ' not expected ' + str(exp_best_path)

        for dir, edge, exp_best_path in [
            (False, ('Shanghai', 'Tokyo'), ['Shanghai', 'Tokyo']),
            (False, ('Shanghai', 'Moscow'), ['Shanghai', 'Tokyo', 'Moscow']),
            (True, ('Shanghai', 'Tokyo'), ['Shanghai', 'Tokyo']),
            (True, ('Shanghai', 'Moscow'), None),
        ]:
            observed_edge = paths_shortest[dir].get(edge, None)
            assert observed_edge == exp_best_path, \
                'Edge ' + str(edge) + ' path ' + str(observed_edge) + ' not expected ' + str(exp_best_path)

        print('-----------------------------------------------------------------------------')
        for dir, source, target, method, exp_path in [
            (False, 'Tokyo', 'Shanghai', 'dijkstra', ['Tokyo', 'Beijing', 'Shanghai']),
            (False, 'Tokyo', 'Shanghai', 'shortest', ['Tokyo', 'Shanghai']),
            (False, 'Tokyo', 'Shanghai', 'simple', ['Tokyo', 'Shanghai']),
            (False, 'Tokyo', 'Antartica', 'dijkstra', None),
            (False, 'Tokyo', 'Antartica', 'shortest', None),
            (False, 'Tokyo', 'Antartica', 'simple', None),
        ]:
            print(str(source) + ' --> ' + str(target))
            paths = gu.get_paths(
                G = G_test[dir],
                source = source,
                target = target,
                method = method,
            )
            print('Best path method "' + str(method) + '" ' + str(source) + '--' + str(target) + ': ' + str(paths))
            best_path = paths[0]['path'] if len(paths)>0 else None
            assert best_path == exp_path, \
                'Best path "' + str(method) + '" ' + str(best_path) + ' not ' + str(exp_path)

        G = gu.create_multi_graph(
            edges = [
                {'key': 'plane', 'u': 'Shanghai', 'v': 'Tokyo', 'distance': 10},
                {'key': 'ship', 'u': 'Shanghai', 'v': 'Tokyo', 'distance': 100},
                {'key': 'plane', 'u': 'Tokyo', 'v': 'Shanghai', 'distance': 22},
                {'key': 'plane', 'u': 'Tokyo', 'v': 'Seoul', 'distance': 5},
                {'key': 'plane', 'u': 'Seoul', 'v': 'Tokyo', 'distance': 6},
                {'key': 'ship', 'u': 'Seoul', 'v': 'Tokyo', 'distance': 60},
            ],
            col_weight = 'distance',
            directed = True,
        )
        for u, v, method, agg, exp_path, exp_weight in [
            ('Shanghai', 'Seoul', 'dijkstra', 'min', ['Shanghai', 'Tokyo', 'Seoul'], 15),
            ('Shanghai', 'Seoul', 'simple', 'min', ['Shanghai', 'Tokyo', 'Seoul'], 15),
            ('Shanghai', 'Seoul', 'simple', 'max', ['Shanghai', 'Tokyo', 'Seoul'], 105),
            # Reverse trip
            ('Seoul', 'Shanghai', 'dijkstra', 'min', ['Seoul', 'Tokyo', 'Shanghai'], 28),
            ('Seoul', 'Shanghai', 'simple', 'min', ['Seoul', 'Tokyo', 'Shanghai'], 28),
            ('Seoul', 'Shanghai', 'simple', 'max', ['Seoul', 'Tokyo', 'Shanghai'], 82),
        ]:
            res = gu.get_paths(G=G, source=u, target=v, method=method, agg_weight_by=agg)
            best_path, best_weight = res[0]['path'], res[0]['weight_total']
            assert best_path == exp_path, \
                'Best path "' + str(method) + '" ' + str(best_path) + ' not ' + str(exp_path)
            assert best_weight == exp_weight, \
                'Best weight "' + str(method) + '" ' + str(best_weight) + ' not ' + str(exp_weight)

        # ---------------------------------------------------------------------------------------------- #

        #
        # Search test
        #
        edge_data_search = [
            {'key': 'plane', 'u': 'Shanghai', 'v': 'Tokyo', 'cost': 10, 'comment': 'Shanghai-Tokyo flight'},
            # duplicate (will not be added), order does not matter
            {'key': 'plane', 'u': 'Tokyo', 'v': 'Shanghai', 'cost': 22, 'comment': 'Tokyo-Shanghai flight'},
            # teleport path Tokyo --> Beijing --> Shanghai shorter distance than plane Tokyo --> Shanghai
            {'key': 'teleport', 'u': 'Tokyo', 'v': 'Beijing', 'cost': 2, 'comment': 'Tokyo-Beijing teleport'},
            {'key': 'teleport-2', 'u': 'Tokyo', 'v': 'Beijing', 'cost': 2, 'comment': 'Tokyo-Beijing teleport'},
            {'key': 'plane', 'u': 'Tokyo', 'v': 'Beijing', 'cost': 9, 'comment': 'Tokyo-Beijing plane'},
            {'key': 'teleport', 'u': 'Beijing', 'v': 'Shanghai', 'cost': 1, 'comment': 'Beijing-Shanghai teleport'},
            # car
            {'key': 'car', 'u': 'Beijing', 'v': 'Moscow', 'cost': 1000000, 'comment': 'Beijing-Moscow car'},
            {'key': 'ev-car', 'u': 'Beijing', 'v': 'Moscow', 'cost': 100000, 'comment': 'Beijing-Moscow EV car'},
            # Other paths
            {'key': 'ship', 'u': 'Shanghai', 'v': 'Xabarovsk', 'cost': 100, 'comment': 'Shanghai-Xabarovsk sea'},
            {'key': 'plane', 'u': 'Xabarovsk', 'v': 'Moscow', 'cost': 3, 'comment': 'Xabarovsk-Moscow plane'},
            {'key': 'plane', 'u': 'Xabarovsk', 'v': 'Tokyo', 'cost': 10, 'comment': 'Asia-Russia flight'},
            {'key': 'train', 'u': 'Xabarovsk', 'v': 'Tokyo', 'cost': 10000, 'comment': 'Asia-Russia train'},
            {'key': 'ship', 'u': 'Xabarovsk', 'v': 'Tokyo', 'cost': MAX_DIST, 'comment': 'Asia-Russia sea'},
            {'key': 'plane', 'u': 'Medellin', 'v': 'Antartica', 'cost': 888, 'comment': 'Medellin-Antartica'},
        ]
        G_search = {}
        for directed in [True, False]:
            G_search[directed] = gu.create_multi_graph(
                edges = edge_data_search,
                col_weight = 'cost',
                directed = directed,
            )
        for i, (dir, query_conns, path_method, exp_top_keys, exp_top_keys_agg_w) in enumerate([
            #
            # Dijkstra test
            #
            (
                    # Undirected test, dijkstra
                    False, [
                        # Among all these connections, we want to find shortest path keys
                        {'u': 'Moscow', 'v': 'Beijing'}, {'u': 'Tokyo', 'v': 'Shanghai'},
                        {'u': 'Medellin', 'v': 'Antartica'}, {'u': 'Vientiane', 'v': 'Bangkok'},
                    ], 'dijkstra',
                    {1.0: ['plane'], 2.0: ['teleport'], 3.0: ['plane', 'teleport'], np.inf: [None]},
                    [
                        {'src_tgt': ('Moscow', 'Beijing'), 'leg_key': 'teleport', '__weight': 0.267},
                        {'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'teleport', '__weight': 1.667},
                        {'src_tgt': ('Moscow', 'Beijing'), 'leg_key': 'plane', '__weight': 7.267},
                        {'src_tgt': ('Medellin', 'Antartica'), 'leg_key': 'plane', '__weight': 888.0},
                    ],
            ),
            (
                    # Directed test, dijkstra
                    True, [
                        {'u': 'Moscow', 'v': 'Beijing'}, {'u': 'Tokyo', 'v': 'Shanghai'},
                        {'u': 'Medellin', 'v': 'Antartica'}, {'u': 'Vientiane', 'v': 'Bangkok'},
                    ], 'dijkstra',
                    {1.0: ['plane'], 2.0: ['teleport'], np.inf: [None]},
                    [
                        {'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'teleport', '__weight': 1.667},
                        {'src_tgt': ('Medellin', 'Antartica'), 'leg_key': 'plane', '__weight': 888.0},
                    ],
            ),
            (
                    # Undirected test, shortest path, by min weight
                    False, [
                        {'u': 'Moscow', 'v': 'Beijing'}, {'u': 'Tokyo', 'v': 'Shanghai'},
                        {'u': 'Medellin', 'v': 'Antartica'}, {'u': 'Vientiane', 'v': 'Bangkok'},
                    ], 'shortest///min',
                    {1.0: ['ev-car', 'plane'], np.inf: [None]},
                    [
                        {'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'plane', '__weight': 22.0},
                        {'src_tgt': ('Medellin', 'Antartica'), 'leg_key': 'plane', '__weight': 888.0},
                        {'src_tgt': ('Moscow', 'Beijing'), 'leg_key': 'ev-car', '__weight': 100000.0},
                    ],
            ),
            (
                    # Undirected test, shortest path, by max weight
                    False, [
                        {'u': 'Moscow', 'v': 'Beijing'}, {'u': 'Tokyo', 'v': 'Shanghai'},
                        {'u': 'Medellin', 'v': 'Antartica'}, {'u': 'Vientiane', 'v': 'Bangkok'},
                    ], 'shortest///max',
                    {1.0: ['car', 'plane'], np.inf: [None]},
                    [
                        {'src_tgt': ('Moscow', 'Beijing'), 'leg_key': 'car', '__weight': 1000000.0},
                        {'src_tgt': ('Medellin', 'Antartica'), 'leg_key': 'plane', '__weight': 888.0},
                        {'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'plane', '__weight': 22.0},
                    ],
            ),
            (
                    # Directed test, shortest path, by max weight
                    True, [
                        {'u': 'Moscow', 'v': 'Beijing'}, {'u': 'Tokyo', 'v': 'Shanghai'},
                        {'u': 'Medellin', 'v': 'Antartica'}, {'u': 'Vientiane', 'v': 'Bangkok'},
                    ], 'shortest///max',
                    {1.0: ['plane'], np.inf: [None]},
                    [
                        {'src_tgt': ('Medellin', 'Antartica'), 'leg_key': 'plane', '__weight': 888.0},
                        {'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'plane', '__weight': 22.0},
                    ],
            ),
            (
                    False, [{'u': 'Bangkok', 'v': 'Moscow'}, {'u': 'Moscow', 'v': 'Shanghai'}], 'dijkstra',
                    {4.0: ['plane', 'teleport'], np.inf: [None]},
                    [
                        {'src_tgt': ('Moscow', 'Shanghai'), 'leg_key': 'teleport', '__weight': 0.312},
                        {'src_tgt': ('Moscow', 'Shanghai'), 'leg_key': 'plane', '__weight': 6.812},
                    ],
            ),
            (
                    False, [{'u': 'Antartica', 'v': 'Medellin'}, {'u': 'Beijing', 'v': 'Shanghai'}], 'dijkstra',
                    {1: ['plane', 'teleport']},
                    [
                        {'src_tgt': ('Beijing', 'Shanghai'), 'leg_key': 'teleport', '__weight': 1.0},
                        {'src_tgt': ('Antartica', 'Medellin'), 'leg_key': 'plane', '__weight': 888.0},
                    ],
            ),
            #
            # Simple test
            #
            (
                    False, [{'u': 'Bangkok', 'v': 'Moscow'}, {'u': 'Tokyo', 'v': 'Shanghai'}],
                    'simple///min',
                    {2: ['teleport'], np.inf: [None]},
                    [{'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'teleport', '__weight': 1.667}],
            ),
            (
                    False, [{'u': 'Bangkok', 'v': 'Moscow'}, {'u': 'Tokyo', 'v': 'Shanghai'}],
                    'simple///max',
                    {4.0: ['car', 'plane', 'ship', 'teleport'], np.inf: [None]},
                    [
                        {'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'car', '__weight': 499999.25},
                        {'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'ship', '__weight': 499998.25},
                        {'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'plane', '__weight': 0.0},
                        {'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'teleport', '__weight': 0.0},
                    ],
            ),
            # Simple graph with given query distance that is closer to 'teleport'
            (
                    False, [{'u': 'Bangkok', 'v': 'Moscow'}, {'u': 'Tokyo', 'v': 'Shanghai', 'cost': 1}],
                    'simple///min',
                    {2: ['teleport'], np.inf: [None]},
                    [{'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'teleport', '__weight': 1.667}],
            ),
            # Simple graph with given query distance that is closer to 'plane'
            (
                    False, [{'u': 'Bangkok', 'v': 'Moscow'}, {'u': 'Tokyo', 'v': 'Shanghai', 'cost': 30}],
                    'simple///min',
                    {1: ['plane'], np.inf: [None]},
                    [{'src_tgt': ('Tokyo', 'Shanghai'), 'leg_key': 'plane', '__weight': 22.0}],
            ),
        ]):
            self.logger.info('Test #' + str(i) + ': ' + str(query_conns))
            if path_method != 'dijkstra':
                path_method, path_agg_wgt_by = path_method.split("///")
            else:
                path_method, path_agg_wgt_by = "dijkstra", "min"

            res = gu.search_top_keys_for_edges(
                query_edges = query_conns,
                ref_multigraph = G_search[dir],
                path_method = path_method,
                # For dijkstra is always "min", so no effect for dijkstra
                path_agg_weight_by = path_agg_wgt_by,
                query_col_u = 'u',
                query_col_v= 'v',
                query_col_weight = 'cost',
                randomize_edge_if_weights_same = False,
            )
            self.logger.info('Return search result:')
            [self.logger.info(str(k) + ': ' + str(v)) for k, v in res.items()]
            top_keys = res['top_keys_by_number_of_edges']
            top_keys_by_agg_weight = res['top_keys_by_aggregated_weight']
            self.logger.info('Top keys ' + str(top_keys) + ', by agg weight ' + str(top_keys_by_agg_weight))
            assert top_keys == exp_top_keys, \
                'Result for test #' + str(i) + ' top keys\n' + str(top_keys) + ' not\n' + str(exp_top_keys)
            assert top_keys_by_agg_weight == exp_top_keys_agg_w, \
                'Result for test #' + str(i) + ' top keys by agg weight\n' \
                + str(top_keys_by_agg_weight) + ' not\n' + str(exp_top_keys_agg_w)

        # gu.draw_graph(G=G_test[False], weight_large_thr=50, agg_weight_by='min')
        self.logger.info('ALL TESTS PASSED')
        return


if __name__ == '__main__':
    Pandas.increase_display()
    lgr = Logging.get_default_logger(log_level=logging.DEBUG, propagate=False)
    GraphUtilsUnitTest(logger=lgr).test()
    exit(0)



import numpy as np
from fitxf.math.datasource.vecdb.model.ModelEncoderInterface import ModelEncoderInterface
from fitxf.math.fit.transform.FitXformPca import FitXformPca
from fitxf.math.utils.Profile import Profiling


class LangModelInterface(ModelEncoderInterface):

    RANDOM_TEXT = \
"""2018년 10월 JYP엔터테인먼트와 정식 매니지먼트 계약을 맺었었다. 이후, 2019년 8월 7일 아티스트컴퍼니와 공식 매니지먼트 계약을 맺었다.
드라마 마녀의 법정, 배드파파, 나쁜 형사에 연달아 출연, 연기 스펙트럼을 쌓아 왔다. 특히, 영화 변신(2019)에서 사춘기 고등학생의 심리를
실감나게 표현해 일상 연기부터 섬뜩한 악마 연기까지 완벽 소화하여 호평받았다. 또한, 드라마 나의 나라(2019)에서는 서휘(양세종)의 동생으로
남선호(우도환)를 좋아하는 서연 역으로 출연하여 첫 사극 도전임에도 안정적인 캐릭터 소화력을 보여줘 ‘괴물 신인’으로 떠올랐다.
2020년, 2021년 tvN 드라마 슬기로운 의사생활에 장윤복 역으로 출연했다. 어리바리한 매력이 있는 쌍둥이 홍도와는 다르게 똘똘하고 열정 넘치는
의대생 윤복 역할을 맡았고, 이후 시즌 2에서는 의사 면허 합격 이후 의사 인턴 역할을 맡았다.
2021년 말부터 2022년 초까지 KBS2 드라마 학교 2021에서 진지원 역으로 출연했다.
2022년, 동명의 네이버 웹툰을 원작으로 한 넷플릭스 오리지널 드라마 지금 우리 학교는에서 반장 최남라 역으로 출연했다. 지금 우리 학교는을
통해 앞으로의 행보가 기대되는 배우로 성장하였다. 극중에서 직전작들[8]과 전혀 다른 캐릭터를 연기했는데,[9] 최남라를 잘 표현해내며 배우로서
성장 가능성을 시사하였다.
"""

    def __init__(
            self,
            model_name = None,
            cache_folder = None,
            include_tokenizer = False,
            logger = None,
    ):
        super().__init__(cache_folder=cache_folder, logger=logger)
        self.model_name = model_name
        self.include_tokenizer = include_tokenizer

        # local model path
        self.model_path = None
        return

    def get_model_name(self):
        return self.model_name

    def get_model_path(self):
        return self.model_path

    def encode(
            self,
            content_list,
            # max length has got no meaning
            maxlen = None,
            return_tensors = 'pt',
            # does not apply here since we can't see the tokenization
            return_attnmasks = False,
            params_other = None,
    ):
        raise Exception('Must be implemented in derived class')

    # Language models usually have max number of tokens, which depends on
    # the tokenization model, either WordPiece, BPE, etc.
    # Thus it is impossible to relate that to how many max characters.
    # It also depends on language, this is only for English.
    # Below algorithm is just a rough guess.
    # TODO Separate out random test text into .txt files for all common languages
    def self_discover_max_len(
            self,
            sample_text = RANDOM_TEXT,
            # last n characters just to avoid accidental errors,
            # make sure this number is larger than the longest word
            last_n_true = 50,
    ):
        n = 100
        last_n_true_count = 0
        emb_prev = None
        # When limit of max tokens reached, we assume the embedding no longer changes
        while n < len(sample_text):
            text = sample_text[0:n]
            self.logger.debug('n = ' + str(n) + ', text char len=' + str(n))
            emb, _ = self.encode(
                content_list = [text],
                return_tensors = 'np',
                return_attnmasks = True,
            )

            if emb_prev is not None:
                is_same = np.sum(emb - emb_prev)**2 < 0.000000001
                self.logger.debug('Is embedding same with previous = ' + str(is_same))
                if is_same:
                    last_n_true_count += 1
                    if last_n_true_count >= last_n_true:
                        self.logger.info(
                            'Embedding remained the same for the last ' + str(last_n_true)
                            + ' characters, starting from length ' + str(n-last_n_true) + '.'
                        )
                        return len(text)-last_n_true, text[0:n-last_n_true]
                else:
                    last_n_true_count = 0

            emb_prev = emb
            n += 1
        # result not found
        return len(sample_text), sample_text

    def speed_test(
            self,
            sentences: list,
            min_rounds = 1000,
    ):
        n = 0
        prof = Profiling(logger=self.logger)
        start_time = prof.start()
        self.logger.info('Start speed test for model "' + str(self.model_name) + '" time ' + str(start_time))
        while True:
            if n >= min_rounds:
                break
            for s in sentences:
                self.encode(content_list=[s])
                n += 1

        diftime = prof.stop() - start_time
        diftime_secs = diftime.days * 86400 + diftime.seconds + diftime.microseconds / 1000000
        rps = n / diftime_secs
        self.logger.info(
            'Model speed test "' + str(self.model_name) + '", rps ' + str(round(rps, 4)) + ', done ' + str(n)
            + ' in ' + str(diftime_secs) + 's'
        )
        return rps

    def visualize_embedding(
            self,
            # numpy.ndarray type
            encoding_np,
            # anything to label the encoding
            labels_list,
    ):
        pca = FitXformPca(logger=self.logger)
        pca.create_scatter_plot2d(
            x_transform = pca.fit(X=encoding_np, n_components=2)[FitXformPca.KEY_X_TRANSFORM],
            labels_list = labels_list,
            show = True,
            save_filepath = None,
        )
        return


if __name__ == '__main__':
    print(LangModelInterface.RANDOM_TEXT)
    # Fake data
    x = np.array([
        [3, 16, 20], [2, 18, 21], [4, 17, 19],
        [15, 19, 5], [17, 22, 5], [16, 20, 3],
    ])
    labels = ['dog', 'dog', 'dog', 'cat', 'cat', 'cat']
    obj = LangModelInterface()
    obj.visualize_embedding(encoding_np=x, labels_list=labels)
    exit(0)



import logging
import os
import re
import torch
import threading
import json
import numpy as np
from datetime import datetime
from fitxf import TensorUtils, FitXformInterface
from fitxf.math.algo.encoding.Base64 import Base64
from fitxf.math.datasource.vecdb.model.ModelEncoderInterface import ModelEncoderInterface
from fitxf.math.datasource.vecdb.model.ModelDbInterface import ModelDbInterface
from fitxf.math.datasource.vecdb.metadata.MetadataInterface import MetadataInterface
from fitxf.math.utils.Lock import Lock


class ModelInterface:

    OLD_DATETIME = datetime(year=1999, month=9, day=9)

    TYPE_TEXT = 'text'
    TYPE_IMG = 'image'
    TYPE_VOICE = 'voice'
    TYPE_SOUND = 'sound'
    TYPE_VIDEO = 'video'

    def __init__(
            self,
            user_id: str,
            # e.g. {'text': modeltxt, 'image': modelimg, ...}
            llm_model: dict[str, ModelEncoderInterface],
            model_db_class: type(ModelDbInterface),
            model_metadata_class: type(MetadataInterface),
            col_content: str,
            col_content_type: str,
            col_label_user: str,
            col_label_std: str,
            col_embedding: str,
            # <=0 means no check
            feature_len: int,
            numpy_to_b64_for_db: bool,
            fit_xform_model: FitXformInterface,
            cache_tensor_to_file: bool,
            file_temp_dir: str,
            # turn off any fitting or transform if False
            in_plain_or_test_mode: bool,
            # allowed values: "np", "torch"
            return_tensors: str = 'np',
            enable_bg_thread_for_training: bool = False,
            logger = None,
    ):
        self.user_id = user_id
        self.llm_model = llm_model
        self.model_db_class = model_db_class
        self.model_metadata_class = model_metadata_class
        self.col_content = col_content
        self.col_content_type = col_content_type
        self.col_label_user = col_label_user
        self.col_label_standardized = col_label_std
        self.col_embedding = col_embedding
        self.feature_len = feature_len
        self.numpy_to_b64_for_db = numpy_to_b64_for_db
        self.fit_xform_model = fit_xform_model
        self.cache_tensor_to_file = cache_tensor_to_file
        self.file_temp_dir = file_temp_dir
        self.in_plain_or_test_mode = in_plain_or_test_mode
        self.return_tensors = return_tensors
        self.enable_bg_thread_for_training = enable_bg_thread_for_training
        self.logger = logger if logger is not None else logging.getLogger()

        self.logger.info(
            'Model in plain/test mode set to "' + str(self.in_plain_or_test_mode) + '"'
        )

        self.base64_encoder = Base64(logger=self.logger)
        self.tensor_utils = TensorUtils(logger=self.logger)
        self.llm_model_path = {k: m.get_model_path() for k, m in self.llm_model.items()}

        self.model_db = self.__get_model_db(ModelClass=self.model_db_class)
        self.logger.info(
            'DB params "' + str(self.user_id) +'": ' + str(self.model_db.get_db_params().get_db_info())
        )

        self.vec_db_metadata = self.__get_model_metadata(MetadataClass=self.model_metadata_class)

        # Lock only by threads in the same worker
        self.mutex_name_model = 'model'
        self.mutex_name_underlying_db = 'db'
        self.lock_mutexes = Lock(
            # User may also provide their own storage (in the usual case)
            mutex_names = [self.mutex_name_model, self.mutex_name_underlying_db],
            logger = self.logger,
        )

        assert os.path.isdir(self.file_temp_dir), 'Not a directory for tmp caching "' + str(self.file_temp_dir) + '"'
        # Don't use characters like "." because during cleanup cant find
        self.file_name_prefix_for_match_pattern = 'VecDbModelStandard_' + \
                                str(datetime.now().strftime("%Y%m%d_%H%M%S")) + '_' \
                                + str(np.random.rand())[2:]
        self.file_name_prefix = self.file_temp_dir + '/' + self.file_name_prefix_for_match_pattern

        self.last_sync_time_with_underlying_db = self.OLD_DATETIME
        self.logger.info(
            'LLM model path "' + str(self.llm_model_path)
            + '", encode numpy to base 64 in DB = ' + str(self.numpy_to_b64_for_db)
        )

        self.bg_thread = threading.Thread(target=self.run_bg_thread)
        self.bg_thread_sleep_secs = float(os.environ.get("VECDB_BG_SLEEP_SECS", "10."))
        self.clear_memory_secs_inactive = float(os.environ.get("VECDB_CLEAR_MEMORY_SECS_INACTIVE", "30."))
        self.signal_stop_bg_thread = False
        self.logger.info(
            'Set bg thread sleep to ' + str(self.bg_thread_sleep_secs)
            + 's, clear memory inactive time to ' + str(self.clear_memory_secs_inactive) + 's.'
        )
        return

    def __get_model_db(
            self,
            ModelClass: type(ModelDbInterface),
    ) -> ModelDbInterface:
        return ModelClass(
            tablename = self.user_id,
            col_content = self.col_content,
            col_content_type = self.col_content_type,
            col_label_user = self.col_label_user,
            col_label_standardized = self.col_label_standardized,
            col_embedding = self.col_embedding,
            max_records = 999999,
            logger = self.logger,
        )

    def __get_model_metadata(
            self,
            MetadataClass: type(MetadataInterface),
    ) -> MetadataInterface:
        return MetadataClass(
            user_id = self.user_id,
            logger = self.logger,
        )

    def check_model_consistency_with_prev(
            self,
    ):
        # Verify that model name has not changed from previous
        try:
            row = self.vec_db_metadata.get_metadata(identifier='llm')
            self.logger.info('Row from metadata for llm: ' + str(row))
            model_prev = row[MetadataInterface.COL_METADATA_VALUE]
        except Exception as ex:
            self.logger.error('Error getting metadata model: ' + str(ex))
            model_prev = None
            # raise Exception(ex)
        self.logger.info('Previous LLM model from metadata "' + str(model_prev) + '"')

        # csv can return float type "nan" instead of None
        if model_prev is not None:
            name_prev = self.get_model_name_from_path(model_path=model_prev)
            name_cur = self.get_model_name_from_path(model_path=self.llm_model_path)
            assert name_prev == name_cur, \
                'Previous model different "' + str(model_prev) + '" (type "' + str(type(model_prev)) + \
                '") from new "' + str(self.llm_model_path) + '" (type "' + str(type(self.llm_model_path)) \
                + '")'
            self.logger.info(
                'Model consistency check ok, cur/prev name same "' + str(name_cur)
                + '", full path "' + str(self.llm_model_path) + '"'
            )
        else:
            self.logger.warning('Model from metadata returned None')

    def cleanup(self):
        try:
            files_in_tmp_dir = os.listdir(path=self.file_temp_dir)
            self.logger.info('Files in temp dir "' + str(self.file_temp_dir) + '": ' + str(files_in_tmp_dir))
            for f in files_in_tmp_dir:
                if re.match(pattern=str(self.file_name_prefix_for_match_pattern)+".*", string=f):
                    self.logger.info(
                        'Cleanup deleting file "' + str(f) + '" using prefix match pattern "'
                        + str(self.file_name_prefix_for_match_pattern) + '"'
                    )
                    os.remove(path=f)
                else:
                    self.logger.debug(
                        'Cleanup ignoring file "' + str(f) + '" using prefix match pattern "'
                        + str(self.file_name_prefix_for_match_pattern) + '"'
                    )
        except Exception as ex:
            self.logger.error('Error cleaning up temp dir "' + str(self.file_temp_dir) + '": ' + str(ex))
        return

    def stop_threads(self):
        if not self.signal_stop_bg_thread:
            self.signal_stop_bg_thread = True
            self.bg_thread.join()
            self.cleanup()
            self.logger.info('All threads successfully stopped')
        return

    def run_bg_thread(
            self,
    ):
        raise Exception('Must be implemented by derived class')

    def extend_feature_len(
            self,
            x: np.ndarray,
    ):
        # no check whatsoever if set to <= 0
        if self.feature_len <= 0:
            return x

        start_dim = x.ndim
        assert start_dim in [1, 2]

        x_tmp = x if start_dim==2 else np.resize(a=x, new_shape=(1, len(x)))
        assert x_tmp.ndim == 2

        h, l = x_tmp.shape
        if l < self.feature_len:
            x_ret = np.append(x_tmp, np.zeros(shape=(h, self.feature_len - l)), axis=-1)
        elif l > self.feature_len:
            raise Exception('Length of array ' + str(l) + ' longer than preset feature length ' + str(self.feature_len))
        else:
            x_ret = x_tmp
        x_final = x_ret if start_dim==2 else x_ret[0]
        # self.logger.debug('Returning extended vector of shape ' + str(x_final.shape))
        return x_final

    def calc_embedding(
            self,
            content_list,
            content_type,
    ):
        assert self.llm_model is not None, 'LLM model is None'
        content_encode = self.llm_model[content_type].encode(
            content_list = content_list,
            return_tensors = self.return_tensors,
        )
        content_encode_std_size = self.extend_feature_len(x=content_encode)
        embed_len = content_encode_std_size.shape[-1]
        if self.feature_len > 0:
            assert embed_len <= self.feature_len, \
                'Embedding length ' + str(embed_len) + ' cannot exceed given fixed feature len ' + str(self.feature_len)
        self.logger.info(
            'Calculated embeddings for content type "' + str(content_type)
            + '", size ' + str(content_encode_std_size.shape)
        )
        return content_encode_std_size

    def get_model_name_from_path(self, model_path):
        parts = [name for name in str(model_path).strip().split(sep=os.sep) if len(name) > 0]
        self.logger.info('Model path separated as ' + str(parts))
        return parts[-1]

    def update_metadata_model_updated(
            self,
            model_save_b64json_string = None,
    ):
        for id, val in (('model', model_save_b64json_string), ('llm', self.llm_model_path)):
            self.vec_db_metadata.update_metadata_identifier_value(
                identifier = id,
                value = val,
            )
        return

    def update_metadata_db_data_updated(self):
        self.vec_db_metadata.update_metadata_identifier_value(
            identifier = 'lastUpdateTimeDb',
            value = datetime.now().strftime(MetadataInterface.DATETIME_FORMAT),
        )
        return

    def init_data_model(
            self,
            max_tries = 1,
            background = False,
    ):
        raise Exception('Must be implemented by derived class')

    def load_data_model(
            self,
            max_tries = 1,
            background = False,
    ):
        raise Exception('Must be implemented by derived class')

    def reset_data_model__(self):
        self.map_lbl_to_idx = {}
        self.map_idx_to_lbl = {}
        self.text_labels_standardized = np.array([])
        self.data_records = []
        return

    def update_model(
            self,
            force_update = False,
            test_mode = False,
    ):
        raise Exception('Must be implemented by derived class')

    def set_high_urgency_to_sync_with_underlying_db(self):
        self.last_sync_time_with_underlying_db = self.OLD_DATETIME

    def get_data_length(self):
        return len(self.text_labels_standardized)

    def atomic_delete_add(
            self,
            delete_key: str,
            # list of dicts
            records: list[dict],
    ):
        raise Exception('Must be implemented by derived class')

    def add(
            self,
            # list of dicts
            records: list,
    ):
        raise Exception('Must be implemented by derived class')

    def delete(
            self,
            match_phrases,
    ):
        raise Exception('Must be implemented by derived class')

    def predict(
            self,
            text_list_or_embeddings,
            content_type: str = TYPE_TEXT,
            # can be cluster numbers to zoom into
            X_embeddings_local_space: np.ndarray = None,
            labels_local_space: list = None,
            top_k = 5,
            # Instead of just returning the user labels, return full record. Applicable to some models only
            return_full_record = False,
    ):
        raise Exception('Must be implemented by derived class')

    def is_updating_model(self):
        return self.lock_mutexes.is_locked(mutex=self.mutex_name_model)

    ############################################################################################################
    # Underlying DB
    ############################################################################################################

    def sync__db(
            self,
            max_tries = 1,
            background = False,
    ):
        raise Exception('Must be implemented by derived class')

    # Heuristic measure from 0 to 1, based on secs since last load time (proportional),
    # training data size (inversely proportional), etc
    def is_need_sync_db(
            self,
    ):
        # secs_since_last_load = Profiling().get_time_dif_secs(
        #     start = self.last_sync_time_with_underlying_db,
        #     stop = last_updated_time_underlying_db
        # )
        row = self.vec_db_metadata.get_metadata(
            identifier = 'lastUpdateTimeDb',
        )
        if row is None:
            db_last_update_time = None
            need_update = True
        else:
            db_last_update_time = datetime.strptime(
                row[MetadataInterface.COL_METADATA_VALUE], MetadataInterface.DATETIME_FORMAT
            )
            need_update = db_last_update_time > self.last_sync_time_with_underlying_db
        self.logger.info(
            'Need update = ' + str(need_update) + '. Last DB update time "' + str(db_last_update_time)
            + '", last sync time ' + str(self.last_sync_time_with_underlying_db)
        )

        if need_update:
            self.logger.info(
                'Need update. Last DB update time "' + str(db_last_update_time)
                + '", last sync time ' + str(self.last_sync_time_with_underlying_db)
            )
        return need_update

    def is_loading_model_from_underlying_db(self):
        return self.lock_mutexes.is_locked(mutex=self.mutex_name_underlying_db)

    def get_text_encoding_from_db_records(
            self,
            db_records,
    ) -> np.ndarray:
        # at the same time, remove embedding record from db_records
        text_encoded = [r.pop(self.col_embedding) for r in db_records]
        if self.numpy_to_b64_for_db:
            text_encoded = [
                # 3 steps: Encode from base 64 string to base 64 bytes, then decode the base 64 to actual bytes,
                # then finally convert to numpy from base 64 bytes
                self.base64_encoder.decode_base64_string_to_numpy_array(s64=s64, data_type='float64')
                for s64 in text_encoded
            ]
        if len(text_encoded) > 0:
            array_lengths = [len(v) for v in text_encoded]
            self.logger.info('Encoding lengths ' + str(array_lengths))
            max_l, min_l = np.max(array_lengths), np.min(array_lengths)
            if self.feature_len > 0:
                assert self.feature_len >= max_l
                if (self.feature_len != max_l) or (max_l != min_l):
                    text_encoded = [self.extend_feature_len(x=v) for v in text_encoded]
                    self.logger.warning(
                        'Appended all vectors to be same length ' + str(max_l) + ', array lengths now '
                        + str([len(v) for v in text_encoded])
                    )
        return np.array(text_encoded)

    def convert_to_embeddings_if_necessary(
            self,
            text_list_or_embeddings,
            content_type,
    ):
        txt_lm = None
        if type(text_list_or_embeddings) is np.ndarray:
            if text_list_or_embeddings.dtype in [
                'float64', 'float32', 'float', 'int64', 'int32', 'int',
                torch.float64, torch.float32, torch.float, torch.int64, torch.int32, torch.int,
            ]:
                txt_lm = text_list_or_embeddings
            else:
                self.logger.info(
                    'Passed in data type "' + str(text_list_or_embeddings.dtype) + '", assuming is text type'
                )

        if txt_lm is None:
            self.logger.info(
                'Passed in data is likely not embedding but text list, calculating embedding: '
                + str(text_list_or_embeddings)
            )
            txt_lm = self.calc_embedding(
                content_list = text_list_or_embeddings,
                content_type = content_type,
            )
        return txt_lm

    def __update_label_mapping(
            self,
            # when new data added, we might need to create new label mapping
            labels_from_new_records,
    ):
        for lb in labels_from_new_records:
            if lb not in self.map_lbl_to_idx:
                while True:
                    int_tmp = np.random.randint(low=0, high=2**31)
                    if int_tmp in self.map_idx_to_lbl.keys():
                        continue
                    self.map_lbl_to_idx[lb] = int_tmp
                    break
        self.map_idx_to_lbl = {i: lbl for lbl, i in self.map_lbl_to_idx.items()}
        return

    def update_label_maps_from_new_recs__(
            self,
            records,
            text_encoding_tensor,
    ):
        labels = [row[self.col_label_user] for row in records]
        self.__update_label_mapping(labels_from_new_records=labels)

        # Save records in append mode
        records_with_embedding = []
        for rec, txt_enc in zip(records, text_encoding_tensor):
            lbl_user = rec[self.col_label_user]
            # add standardized label to existing object in passed in records
            rec[self.col_label_standardized] = self.map_lbl_to_idx.get(lbl_user, -1)

            # Create a new record to be appended, don't add embedding to passed in records
            rec_copy = {k:v for k,v in rec.items()}
            rec_copy[self.col_embedding] = txt_enc
            records_with_embedding.append(rec_copy)

        return records_with_embedding

    def add_records_to_underlying_db__(
            self,
            records_with_embedding_and_labelstd,
    ):
        if self.numpy_to_b64_for_db:
            recs_for_storage_base64 = []
            for rec in records_with_embedding_and_labelstd:
                r = rec.copy()
                r[self.col_embedding] = self.base64_encoder.encode_numpy_array_to_base64_string(
                    x = r[self.col_embedding],
                    data_type = 'float64',
                )
                recs_for_storage_base64.append(r)
            self.logger.info(
                'Successfully encoded all embedding to base 64 string for storage.'
            )
        else:
            recs_for_storage_base64 = records_with_embedding_and_labelstd

        self.model_db.insert(
            records = recs_for_storage_base64,
            tablename = self.user_id,
        )
        self.logger.info(
            'Train update underlying DB saved ' + str(len(records_with_embedding_and_labelstd)) + ' records: '
            + str([{k: v for k, v in d.items() if k != self.col_embedding} for d in recs_for_storage_base64])
        )
        return

    def delete_records_from_underlying_db__(
            self,
            match_phrases,
    ):
        total_deleted = 0
        # Get all encoded texts, standardized labels, text encoding from memory cache
        for mp in match_phrases:
            self.__delete_single_record_from_underlying_db(
                match_phrase = mp,
            )
        return total_deleted

    def __delete_single_record_from_underlying_db(
            self,
            match_phrase,
    ):
        total_deleted = 0
        assert len(match_phrase) == 1, 'Not supported delete with more than 1 key phrase ' + str(match_phrase)
        self.logger.info('Trying to delete records using match phrase: ' + str(match_phrase))
        res_del = self.model_db.delete(
            match_phrase = match_phrase,
            tablename = self.user_id,
        )
        self.logger.info('Deleted record with match phrase ' + str(match_phrase) + ', result ' + str(res_del))
        return 1


if __name__ == '__main__':
    exit(0)



import pandas as pd


class Pandas:

    @staticmethod
    def increase_display(
            display_max_rows = 500,
            display_max_cols = 500,
            display_width    = 1000,
    ):
        pd.set_option('display.max_rows', display_max_rows)
        pd.set_option('display.max_columns', display_max_cols)
        pd.set_option('display.width', display_width)


# -*- coding: utf-8 -*-
import flask
import logging
import os
from fitxf.math.utils.CmdLine import CmdLine


#
# In order to make gunicorn work, make sure in your derived class to add this line
#    from ub.api_rest.RestFlaskMultiThread import app as application
# This is because gunicorn will look for the object "application" to call methods to
# start the Flask server
#
app = flask.Flask(__name__)

class RestFlaskMultiThread:

    JSON_ENCODING = 'utf-8'

    def __init__(
            self,
            port = 80,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        # Flask app
        self.app = app
        # self.app_test.config['DEBUG'] = False
        self.params_cl = self.get_cmdline_params(default_port=port)
        self.port = self.params_cl['port']
        self.logger.info('Port from commend line "' + str(self.port) + '"')
        self.init_rest_urls()
        return

    def get_cmdline_params(self, default_port):
        params = CmdLine.get_cmdline_params()
        for p, p_env_var, v_default in [('port', 'PORT', default_port)]:
            if p not in params.keys():
                # not in command line, try env var instead
                v_envvar_or_default = os.environ.get(p_env_var, v_default)
                params[p] = v_envvar_or_default
                self.logger.warning(
                    'Param "' + str(p) + '" not in command line, default to value from env var or default "'
                    + str(v_envvar_or_default) + '"'
                )
        return params

    def init_rest_urls(self):
        raise Exception('Must be implemented by derived class')
        # Example http://localhost:7777/?text=I%20love%20Vader%20Sentiment
        # @self.app_test.route('/', methods=['POST','GET'])
        # def call():
        #     request_json = self.get_request_json(method=flask.request.method)
        #     m = flask.request.method
        #     # Do something
        #     return json.dumps({
        #         'text': text,
        #         'sentiment': res,
        #         'req': str(request_json),
        #     })

    def get_request_json(
            self,
            method = None
    ):
        method = flask.request.method if method is None else method
        req_json = {}
        if method == 'GET':
            for param_name in flask.request.args:
                req_json[param_name] = flask.request.args[param_name]
            return req_json
        else:
            #obj = json.loads(flask.request.json, encoding=AggregateServer.JSON_ENCODING)
            return flask.request.json

    def get_param(self, param_name, method=None):
        method = flask.request.method if method is None else method
        if method == 'GET':
            if param_name in flask.request.args:
                return str(flask.request.args[param_name])
            else:
                return None
        else:
            try:
                val = flask.request.json[param_name]
                return val
            except Exception as ex:
                logging.critical(
                    str(self.__class__) + ': No param name "' + str(param_name) + '" in request: ' + str(ex)
                )
                return None

    def get_params(self, param_list, method=None):
        method = flask.request.method if method is None else method
        data = {}
        for p in param_list:
            data[p] = self.get_param(param_name=p, method=method)
            self.logger.info('Read param "' + str(p) + '": "' + str(data[p]) + '"')

        return data.values()

    def get_remote_ip_from_nginx(
            self,
            key = None,
    ):
        key = 'X-Real-IP' if key is None else key
        x_real_ip = flask.request.headers.get(key, None)
        self.logger.debug('Real IP from key "' + str(key) + '" = "' + str(x_real_ip) + '"')
        return x_real_ip

    def get_remote_port_from_nginx(
            self,
            key = None,
    ):
        key = 'X-Real-Port' if key is None else key
        x_real_port = flask.request.headers.get(key, None)
        self.logger.debug('Real Port from key "' + str(key) + '" = "' + str(x_real_port) + '"')
        return x_real_port

    def run(
            self,
    ):
        try:
            self.app.run(
                host = '0.0.0.0',
                port = self.port
            )
        except Exception as ex:
            self.logger.critical(
                str(self.__class__) + ': Could not start REST Server, got exception: ' + str(ex)
            )

    def stop(
            self
    ):
        self.logger.critical(
            str(self.__class__) + ': Stop aggregate server command received.'
        )
        # TODO Stop server
        func = flask.request.environ.get('werkzeug.server.shutdown')
        if func is None:
            raise RuntimeError('Not running with the Werkzeug Server')
        func()
        return


"""
Example class to derive from this RestFlaskMultiThread class.
Then instantiate the class outside of main for gunicorn.
"""
# class ExampleRest(RestFlaskMultiThread):
#     def __init__(self):
#         super().__init__()
#
#     def init_rest_urls(self):
#         # Example http://localhost:7777/?text=yo
#         @self.app.route('/', methods=['POST','GET'])
#         def call():
#             request_json = self.get_request_json(method=flask.request.method)
#             # Do something
#             return json.dumps({
#                 'datetime': datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
#                 'port': self.port,
#                 'request': str(request_json),
#             })
# __inst = ExampleRest()

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    # If running from gunicorn, no need to run this (will be started by
    # gunicorn instead with host/port given to gunicorn on the command line)
    #__inst.run()



import logging
import torch
from fitxf.math.fit.arc.ClassifierArcInterface import ClassifierArcInterface
from fitxf.math.fit.arc.ClassifierArcUnitTest import ClassifierArcUnitTest
from fitxf import FitUtils
from fitxf.math.utils.Logging import Logging


class ClassifierArc(torch.nn.Module, ClassifierArcInterface):

    # Problem with using dynamic array to store layers is that we can't load back from file (torch limitation)
    USE_ARRAY = False

    ACT_FUNCS = (
        torch.nn.Tanh, torch.nn.ReLU, torch.nn.Sigmoid, torch.nn.LogSigmoid, torch.nn.Softmax,
    )

    def __init__(
            self,
            model_filepath: str = None,
            in_features: int = None,
            out_features: int = None,
            n_hidden_features: int = 100,
            hidden_functions: list = (torch.nn.Linear, torch.nn.Linear, torch.nn.Linear),
            activation_functions: list = (torch.nn.ReLU, torch.nn.ReLU, torch.nn.Softmax),
            loss_function = torch.nn.CrossEntropyLoss,
            dropout_rate: float = 0.2,
            learning_rate: float = 0.0001,
            logger = None,
    ):
        torch.nn.Module.__init__(self)
        ClassifierArcInterface.__init__(
            self = self,
            model_filepath = model_filepath,
            in_features = in_features,
            out_features = out_features,
            n_hidden_features = n_hidden_features,
            hidden_functions = hidden_functions,
            activation_functions = activation_functions,
            loss_function = loss_function,
            dropout_rate = dropout_rate,
            learning_rate = learning_rate,
            logger = logger,
        )

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.__init_neural_network_arc()
        if self.model_filepath is not None:
            # if loading from filepath, will overwrite the above neural network params (if different)
            self.from_old_states(model_filepath=self.model_filepath)
        return

    def __init_neural_network_arc(
            self,
    ):
        assert len(self.hidden_functions) == 3, \
            'Expect 3 hidden functions but got ' + str(len(self.hidden_functions))
        assert self.hidden_functions[0] is not None

        n_out_1 = self.n_hidden_features[0]
        self.layer_hidden_fc1 = self.hidden_functions[0](
            in_features = self.in_features,
            out_features = n_out_1,
        )
        if self.activation_functions[0] in self.ACT_FUNCS:
            self.layer_hidden_fc1_act = self.activation_functions[0]()
        elif self.activation_functions[0] == torch.nn.LayerNorm:
            self.layer_hidden_fc1_act = torch.nn.LayerNorm(
                normalized_shape = n_out_1,
            )
        else:
            self.layer_hidden_fc1_act = None
            self.logger.warning(
                'Ignore non-recognized activation function "'
                + str(self.activation_functions[0]) + '"'
            )
        self.layer_dropout_fc1_drop = torch.nn.Dropout(p=self.dropout_rate)

        self.layer_hidden_fc2 = None
        self.layer_hidden_fc2_act = None
        self.layer_dropout_fc2_drop = None
        if self.hidden_functions[1] is not None:
            n_out_2 = self.n_hidden_features[1]
            self.layer_hidden_fc2 = self.hidden_functions[1](
                in_features = n_out_1,
                out_features = n_out_2,
            )
            if self.activation_functions[1] in self.ACT_FUNCS:
                self.layer_hidden_fc2_act = self.activation_functions[1]()
            elif self.activation_functions[1] == torch.nn.LayerNorm:
                self.layer_hidden_fc2_act = torch.nn.LayerNorm(
                    normalized_shape = n_out_2,
                )
            else:
                self.logger.warning(
                    'Ignore non-recognized activation function "'
                    + str(self.activation_functions[0]) + '"'
                )
            self.layer_dropout_fc2_drop = torch.nn.Dropout(p=self.dropout_rate)
        else:
            n_out_2 = n_out_1

        self.layer_hidden_last = None
        self.layer_act_last = None
        if self.hidden_functions[2] is not None:
            self.layer_hidden_last = torch.nn.Linear(
                in_features = n_out_2,
                out_features = self.out_features,
            )
            if self.activation_functions[2] in self.ACT_FUNCS:
                self.layer_act_last = self.activation_functions[2]()
            elif self.activation_functions[2] == torch.nn.LayerNorm:
                self.layer_act_last = torch.nn.LayerNorm(
                    normalized_shape = self.out_features,
                )
            else:
                self.logger.warning(
                    'Ignore non-recognized activation function "'
                    + str(self.activation_functions[2]) + '"'
                )

        # Random initialization of model parameters if not loading from a previous state
        for p in self.parameters():
            if p.dim() > 1:
                torch.nn.init.xavier_uniform_(p)
        self.loss_func = self.loss_function()
        self.optimizer = torch.optim.Adam(
            self.parameters(),
            lr = self.learning_rate,
            betas = (0.9, 0.98),
            eps = 1e-9
        )
        self.logger.info('Network initialized successfully ' + str(self))
        return

    def from_old_states(
            self,
            model_filepath: str,
    ):
        state = torch.load(
            f = model_filepath,
            map_location = self.device,
        )
        model_state_dict = state['model_state_dict']
        optimizer_state_dict = state['optimizer_state_dict']
        self.load_state_dict(
            state_dict = model_state_dict,
        )
        self.optimizer.load_state_dict(
            state_dict = optimizer_state_dict
        )
        self.logger.info('Loaded old state from file "' + str(model_filepath) + '"')
        return state

    def save_states(
            self,
            model_filepath: str,
            additional_info: dict,
    ):
        state = {
            'model_state_dict': self.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
        }
        state.update(additional_info)
        torch.save(obj=state, f=model_filepath)
        self.logger.info('Saved state dicts for model/optimizer to "' + str(model_filepath) + '"')

    def forward(
            self,
            x: torch.Tensor,
    ):
        h1 = self.layer_hidden_fc1(x)
        # self.logger.debug('Linear layer shape ' + str(h1.shape))
        if self.layer_hidden_fc1_act is not None:
            h1_act_or_norm = self.layer_hidden_fc1_act(h1)
        else:
            h1_act_or_norm = h1
        h1_out = self.layer_dropout_fc1_drop(h1_act_or_norm)

        if self.layer_hidden_fc2 is not None:
            h2 = self.layer_hidden_fc2(h1_out)
            if self.layer_hidden_fc2_act is not None:
                h2_act_or_norm = self.layer_hidden_fc2_act(h2)
            else:
                h2_act_or_norm = h2
            h2_out = self.layer_dropout_fc2_drop(h2_act_or_norm)
        else:
            h2_out = h1_out

        if self.layer_hidden_last is not None:
            h_last = self.layer_hidden_last(h2_out)
            # self.logger.debug('Sentiment linear layer shape ' + str(senti.shape))
            if self.layer_act_last is not None:
                last_out = self.layer_act_last(h_last)
            else:
                last_out = h_last
        else:
            last_out = h2_out
        return last_out

    def fit(
            self,
            X: torch.Tensor,
            y: torch.Tensor,
            is_categorical: bool = True,
            # if None, means we don't convert to onehot (possibly caller already done that, or not required)
            num_categories: int = None,
            # the smaller the batch size, the smaller the losses will be during training
            batch_size: int = 32,
            epochs: int = 100,
            eval_percent: float = 0.2,
            # important to prevent over-fitting
            regularization_type = "L2",
    ):
        fit_utils = FitUtils(logger=self.logger)
        dl_trn, dl_val, n_cutoff_train = self.create_dataloader(
            X = X,
            y = y,
            y_num_classes = num_categories,
            batch_size = batch_size,
            eval_percent = eval_percent,
            include_attn_mask = True,
        )
        self.logger.info(
            'Train length = ' + str(len(dl_trn)) + ', eval length = ' + str(len(dl_val))
            + ', cutoff train = ' +  str(n_cutoff_train)
        )

        losses = fit_utils.torch_train(
            model = self,
            train_dataloader = dl_trn,
            loss_func = self.loss_func,
            optimizer = self.optimizer,
            regularization_type = regularization_type,
            epochs = epochs,
        )
        self.logger.info('Train losses: ' + str(losses))
        # Important! Set back to eval mode, so subsequent forwards won't affect any weights or gradients
        self.eval()

        if eval_percent > 0:
            eval_accuracy = self.evaluate_accuracy(
                X_eval = X[n_cutoff_train:],
                y_eval = y[n_cutoff_train:],
                is_categorical = is_categorical,
                num_categories = num_categories,
            )
        else:
            eval_accuracy = None

        return {
            'eval_accuracy': eval_accuracy,
            'losses': losses,
            'dataloader_train': dl_trn,
            'dataloader_eval': dl_val,
            'index_cutoff_train': n_cutoff_train,
        }

    def evaluate_accuracy(
            self,
            X_eval: torch.Tensor,
            y_eval: torch.Tensor,
            is_categorical: bool = True,
            # if None, means we don't convert to onehot (possibly caller already done that, or not required)
            num_categories: int = None,
    ):
        assert len(X_eval) == len(y_eval)
        out_sorted, _ = self.predict(X=X_eval)
        self.logger.debug(
            'Out for eval test (shape ' + str(out_sorted.shape) + ', y shape ' + str(y_eval.shape)
            + '): ' + str(out_sorted)
        )
        out_top = out_sorted[:,0]
        y_expected = y_eval

        self.logger.debug(
            'Out categories for eval test:\n' + str(list(zip(out_top.tolist(), y_expected.tolist())))
        )
        correct = 1 * (y_expected - out_top == 0)
        eval_accuracy = torch.sum(correct) / len(correct)
        eval_accuracy = eval_accuracy.item()
        self.logger.info(
            'Evaluation results: Total correct ' + str(torch.sum(correct).item()) + ' from length ' + str(len(correct))
            + ', accuracy ' + str(eval_accuracy)
        )
        return eval_accuracy

    def predict(
            self,
            X: torch.Tensor,
    ):
        self.eval()
        out = self(X)
        # self.logger.debug('Predict input X:\n' + str(X[:,:8]) + '\nOut for predict:\n' + str(out))
        out_arg_sorted = torch.argsort(out, descending=True, dim=-1)
        # out_cat = torch.argmax(out, dim=-1)
        self.logger.debug('Out arg sorted for predict: ' + str(out_arg_sorted))
        assert len(out) == len(out_arg_sorted)
        out_val_sorted = torch.Tensor([[out[i][j].item() for j in row] for i, row in enumerate(out_arg_sorted)])
        self.logger.debug('Model output inference argsort: ' + str(out_val_sorted))

        return out_arg_sorted, out_val_sorted


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    ut = ClassifierArcUnitTest(child_class=ClassifierArc, logger=lgr)
    ut.test()
    exit(0)



## Basic Math

import numpy as np
from fitxf import TensorUtils
ts = TensorUtils()
x = np.random.rand(5,3)
y = np.random.rand(10,3)

# Cosine similarity, find closest matches of each vector in x
# with all vectors in ref
# For Euclidean distance, just replace with "similarity_distance"
ts.similarity_cosine(x=x, ref=x, return_tensors='np')
ts.similarity_cosine(x=y, ref=y, return_tensors='np')
matches, dotsim = ts.similarity_cosine(x=x, ref=y, return_tensors='np')
print("matches",matches)
print("dot similarities",dotsim)


## Clustering

from fitxf import Cluster
x = np.array([
    [5, 1, 1], [8, 2, 1], [6, 0, 2],
    [1, 5, 1], [2, 7, 1], [0, 6, 2],
    [1, 1, 5], [2, 1, 8], [0, 2, 6],
])
obj = Cluster()
obj.kmeans_optimal(
    x = x,
    estimate_min_max = True,
)


from fitxf import ClusterCosine
x = np.random.rand(20,3)
ClusterCosine().kmeans_optimal(x=x)


## Graph Wrappers

from fitxf import GraphUtils
gu = GraphUtils()
G = gu.create_multi_graph(
    edges = [
        {'key': 'plane', 'u': 'Shanghai', 'v': 'Tokyo', 'distance': 10},
        {'key': 'ship', 'u': 'Shanghai', 'v': 'Tokyo', 'distance': 100},
        {'key': 'plane', 'u': 'Tokyo', 'v': 'Shanghai', 'distance': 22},
        {'key': 'plane', 'u': 'Tokyo', 'v': 'Seoul', 'distance': 5},
        {'key': 'plane', 'u': 'Seoul', 'v': 'Tokyo', 'distance': 6},
        {'key': 'ship', 'u': 'Seoul', 'v': 'Tokyo', 'distance': 60},
    ],
    col_weight = 'distance',
    directed = True,
)
# Shanghai-->Tokyo-->Seoul, total weight 15
print(gu.get_paths(G=G, source="Shanghai", target="Seoul", method="dijkstra"))
# Shanghai-->Tokyo-->Seoul, total weight 105
print(gu.get_paths(G=G, source="Shanghai", target="Seoul", method="simple", agg_weight_by="max"))
# Seoul-->Tokyo-->Shanghai, total weight 28
print(gu.get_paths(G=G, source="Seoul", target="Shanghai", method="dijkstra"))
# Shanghai-->Tokyo-->Seoul, total weight 82
print(gu.get_paths(G=G, source="Seoul", target="Shanghai", method="simple", agg_weight_by="max"))


## Fit Transform

from fitxf import FitXformPca, FitXformCluster
import numpy as np
x = np.array([
    [5, 1, 1], [8, 2, 1], [6, 0, 2],
    [1, 5, 1], [2, 7, 1], [0, 6, 2],
    [1, 1, 5], [2, 1, 8], [0, 2, 6],
])
user_labels = [
    'a', 'a', 'a',
    'b', 'b', 'b',
    'c', 'c', 'c',
]
pca = FitXformPca()
res_fit_pca = pca.fit_optimal(X=x, X_labels=user_labels)
print('X now reduced to\n',res_fit_pca['X_transform'])

cls = FitXformCluster()
res_fit_cls = cls.fit_optimal(X=x, X_labels=user_labels)
print('X now reduced to\n',res_fit_cls['X_transform'])

pca.predict(X=x+np.random.rand(9,3))
cls.predict(X=x+np.random.rand(9,3))


### Save Model to String & Load Back

# Save this json string somewhere
model_save = pca.model_to_b64json(numpy_to_base64_str=True, dump_to_b64json_str=True)

# Load back into new instance
new = FitXformPca()
new.load_model_from_b64json(model_b64json=model_save)
new.predict(X=x+np.random.rand(9,3))


### Fine Tuning

cls = FitXformCluster()
res_fit_cls = cls.fit_optimal(X=x, X_labels=user_labels)
centers = res_fit_cls["centers"]

res = cls.fine_tune(
   X = x,
   X_labels = user_labels,
   n_components = 3,
)
[print(k,v) for k,v in res.items()]
print('Expect 1 iteration, got ', res["n_iter"])

print('old centers',centers)
print('new centers',res['centers'])


x = np.array([
    [5, 1, 1], [8, 2, 1], [6, 0, 2],
    [1, 5, 1], [2, 7, 1], [0, 6, 2],
    # remove last 2 points
    [1, 1, 5], # [2, 1, 8], [0, 2, 6],
    # new points
    [4, 2, 1], [0, 6, 2], [1, 1, 7],
])
user_labels = [
    'a', 'a', 'a',
    'b', 'b', 'b',
    # remove last 2 points
    'c', # 'c', 'c',
    'a', 'b', 'c',
]
res_fit_new = cls.fine_tune(X=x, X_labels=user_labels, n_components=3)

print('old centers',centers)
print('new centers',res_fit_new['centers'])



from fitxf.math.utils.Env import Env
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Lock import Lock
from fitxf.math.utils.Singleton import Singleton, SingletonMemoryMgmt, SingletonUnitTest
from fitxf.math.utils.LockF import LockFile, LoadTestLockFile
from fitxf.math.utils.ObjPers import ObjectPersistence
from fitxf.math.utils.Profile import Profiling



from fitxf.math.lang.encode.LangModelInterface import LangModelInterface
from fitxf.math.utils.Singleton import Singleton


class LangModelPtSingleton:

    @staticmethod
    def get_singleton(
            LmClass,    # class type
            model_name = None,
            cache_folder = None,
            include_tokenizer = False,
            logger = None,
            return_key = False,
    ) -> LangModelInterface:
        # Don't include lang, because it may change to "multi"
        key_id = str(LmClass) + '.model_name=' + str(model_name) \
                 + '.cache_folder=' + str(cache_folder) + '.include_tokenizer=' + str(include_tokenizer)
        sgt = Singleton(
            class_type = LmClass,
            logger = logger,
        ).get_singleton(
            key_id,
            model_name,
            cache_folder,
            include_tokenizer,
            logger,
        )
        return (sgt, key_id) if return_key else sgt



import logging
from datetime import datetime


class MetadataInterface:

    DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S.%f'

    COL_METADATA_USERID = 'metadata_userid'
    COL_METADATA_IDENTIFIER = 'metadata_identifier'
    COL_METADATA_TIMESTAMP = 'metadata_timestamp'
    COL_METADATA_VALUE = 'metadata_value'

    def __init__(
            self,
            # name to identify which user/table/etc this metadata is referring to
            user_id,
            metadata_tbl_name = 'model_metadata',
            logger = None,
    ):
        self.user_id = user_id
        self.metadata_tbl_name = metadata_tbl_name
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def get_timestamp(
            self,
            dt: datetime = None,
    ):
        dt = datetime.now() if dt is None else dt
        return datetime.timestamp(dt)

    def get_datetime_from_timestamp(
            self,
            timestamp: float,
    ):
        return datetime.fromtimestamp(timestamp)

    def get_metadata(
            self,
            identifier
    ):
        raise Exception('Must be implemented by child class')

    # signify that model has been updated
    def update_metadata_identifier_value(
            self,
            identifier: str,
            value: str,
    ):
        raise Exception('Must be implemented by child class')

    def cleanup(
            self,
    ):
        raise Exception('Must be implemented by child class')


if __name__ == '__main__':
    m = MetadataInterface(user_id='main')
    now = datetime.now()
    ts = m.get_timestamp(dt=now)
    dt = m.get_datetime_from_timestamp(timestamp=ts)
    print(ts, dt, now)
    exit(0)



import warnings
import logging
import threading
import os
import builtins
from fitxf.math.utils.Env import Env
from fitxf.math.utils.Logging import Logging


class DbParams:

    @staticmethod
    def get_db_params_from_envvars(
            identifier,
            db_create_tbl_sql,
            db_table,
            verify_certs = False,
    ):
        return DbParams(
            identifier = identifier,
            db_type = os.environ["DB_TYPE"],
            db_host = os.environ["DB_HOST"],
            db_port = os.environ["DB_PORT"],
            db_scheme = os.environ["DB_SCHEME"],
            db_username = os.environ["DB_USERNAME"],
            db_password = os.environ["DB_PASSWORD"],
            db_database = os.environ["DB_DATABASE"],
            db_root_folder = os.environ["DB_ROOT_FOLDER"],
            db_create_table_sql = db_create_tbl_sql,
            db_table = db_table,
            db_verify_certs = verify_certs,
        )

    def __init__(
            self,
            # any name to indentify this set of DB params
            identifier,
            db_type = None,
            db_host = None,
            db_port = None,
            db_scheme = None,
            db_username = None,
            db_password = None,
            db_database = None,
            db_verify_certs = True,
            db_table = None,
            db_create_table_sql = None,
            # for certain types of DB types like csv
            db_root_folder = None,
            params_other = None,
            logger = None,
    ):
        self.identifier = identifier
        self.db_type = db_type if db_type is not None else os.environ["DB_TYPE"]
        self.db_host = db_host if db_host is not None else os.environ["DB_HOST"]
        self.db_port = db_port if db_port is not None else os.environ["DB_PORT"]
        self.db_scheme = db_scheme if db_scheme is not None else os.environ["DB_SCHEME"]
        self.db_username = db_username if db_username is not None else os.environ["DB_USERNAME"]
        self.db_password = db_password if db_password is not None else os.environ["DB_PASSWORD"]
        self.db_database = db_database if db_database is not None else os.environ["DB_DATABASE"]
        self.db_verify_certs = db_verify_certs
        self.db_table = db_table
        self.db_create_table_sql = db_create_table_sql
        self.db_root_folder = db_root_folder if db_root_folder is not None else os.environ["DB_ROOT_FOLDER"]
        self.params_other = {} if params_other is None else params_other
        self.logger = logger if logger is not None else logging.getLogger()

        passwd = str(self.db_password)
        hide_len = min(4, int(0.5*len(passwd)))
        mask_len = len(passwd) - hide_len
        mask_len = max(0, mask_len)
        self.db_password_hide = str('*' * mask_len) + str(passwd[-hide_len:])
        self.logger.info('Initialized DB Params as: ' + str(self.get_db_info()))
        return

    def get_db_info(self):
        return \
            'DB Params for "' + str(self.identifier) + '" as follows: ' \
            + 'DB Type "' + str(self.db_type) + '", host "' + str(self.db_host) \
            + '", port ' + str(self.db_port) \
            + ', scheme "' + str(self.db_scheme) \
            + '", username "' + str(self.db_username)\
            + '", password "' + str(self.db_password_hide) \
            + '", database "' + str(self.db_database) \
            + '", verify certs "' + str(self.db_verify_certs) \
            + '", index "' + str(self.db_table) \
            + '", create table sql "' + str(self.db_create_table_sql) \
            + ', folder "' + str(self.db_root_folder) \
            + '", params other: ' + str(self.params_other)

    def copy_db_params_with_new_index(
            self,
            identifier,
            new_table,
    ):
        return DbParams(
            identifier = identifier,
            db_type = self.db_type,
            db_host = self.db_host,
            db_port = self.db_port,
            db_scheme = self.db_scheme,
            db_username = self.db_username,
            db_password = self.db_password,
            db_database = self.db_database,
            db_verify_certs = self.db_verify_certs,
            db_table = new_table,
            db_create_table_sql = self.db_create_table_sql,
            # for certain types of DB types like csv
            db_root_folder = self.db_root_folder,
            params_other = self.params_other,
            logger = self.logger,
        )


class DatastoreInterface:

    def __init__(
            self,
            # connection options, can pass during connection or here
            db_params: DbParams,
            ignore_warnings = False,
            logger = None,
    ):
        if ignore_warnings:
            # Ignore warning "InsecureRequestWarning: Unverified HTTPS request is being made to host 'localhost'..."
            warnings.filterwarnings("ignore")
        self.db_params = db_params
        self.logger = logger if logger is not None else logging.getLogger()

        self.con = None
        self.timeout = int(os.environ.get('DB_DEFAULT_TIMEOUT', 30))
        self.mutex_write = threading.Lock()
        return

    def connect(
            self,
    ):
        raise Exception('Must be implemented by derived class')

    def close(
            self,
    ):
        raise Exception('Must be implemented by derived class')

    def get(
            self,
            # e.g. {"answer": "take_seat"}
            match_phrase,
            match_condition: dict = {'and': True, 'exact': True},
            tablename = None,
            request_timeout = 20.0,
    ):
        raise Exception('Must be implemented by derived class')

    def get_all(
            self,
            key = None,
            max_records = 10000,
            tablename = None,
            request_timeout = 20.0,
    ):
        raise Exception('Must be implemented by derived class')

    def get_indexes(self):
        raise Exception('Must be implemented by derived class')

    def delete_index(
            self,
            tablename,
    ):
        raise Exception('Must be implemented by derived class')

    def get_column_names(
            self,
            tablename,
    ):
        records = self.get_all(
            tablename = tablename,
            max_records = 1,
        )
        assert len(records) > 0, 'Cannot get column names by default method without data for "' + str(tablename) + '"'
        return [k for k in records[0].keys()]

    def get_mapping(
            self,
            tablename = None,
    ):
        raise Exception('Must be implemented by derived class')

    def atomic_delete_add(
            self,
            delete_key: str,
            # list of dicts, python3.8 & below does not support this syntax
            records: list[dict],
            tablename: str = None,
    ):
        raise Exception('Must be implemented by derived class')

    def add(
            self,
            # list of dicts, python3.8 & below does not support this syntax
            records: list[dict],
            tablename: str = None,
    ):
        raise Exception('Must be implemented by derived class')

    def delete(
            self,
            match_phrase: dict,
            match_condition: dict = {'and': True, 'exact': True},
            tablename: str = None,
    ):
        raise Exception('Must be implemented by derived class')

    def delete_by_raw_query(
            self,
            raw_query,
            tablename = None,
    ):
        raise Exception('Must be implemented by derived class')

    def delete_all(
            self,
            tablename = None,
    ):
        raise Exception('Must be implemented by derived class')

    def add_column(
            self,
            colnew,
            data_type = str,
            tablename = None,
            default_value = None,
    ):
        raise Exception('Must be implemented by derived class')

    #
    # Set callback to receive data
    #
    def set_receive_callback(
            self,
            callback,
    ):
        raise Exception('Must be implemented by derived class')


class DatastoreInterfaceUnitTest:

    def __init__(
            self,
            ChildClass: type(DatastoreInterface),
            logger = None,
    ):
        self.ChildClass = ChildClass
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def get_child_class(
            self,
            tablename,
    ) -> DatastoreInterface:
        return self.ChildClass(
            db_params = DbParams.get_db_params_from_envvars(
                identifier = 'testdbif',
                db_create_tbl_sql = '',
                db_table = tablename,
            ),
            logger=self.logger,
        )

    def test(
            self,
            tablename,
    ):
        ds = self.get_child_class(tablename = tablename)
        ds.delete_index(tablename=tablename)
        ds.connect()

        rows = ds.get_all()
        assert len(rows) == 0, 'Should be empty but is not: ' + str(rows)

        new_records = [
            {'id': 100, 'text': "sentence A", 'answer': ' xx'},
            {'id': 101, 'text': "sentence A", 'answer': ' xx'},
            {'id': 102, 'text': "sentence B", 'answer': ' yy'},
            {'id': 103, 'text': "sentence B", 'answer': ' yy'},
            {'id': 103, 'text': "sentENce B", 'answer': ' zz'},
            {'id': 104, 'text': "this is sentence B", 'answer': ' zz'},
        ]
        ds.add(records=new_records)

        colnames = ds.get_column_names(tablename=tablename)
        self.logger.info('Column names derived: ' + str(colnames))
        assert colnames == ['id', 'text', 'answer'], 'Column names ' + str(colnames) + ' not expected'

        rows = ds.get_all()
        assert len(rows) == len(new_records), \
            'Length of memory cache ' + str(len(rows)) + ' not same with data added ' + str(len(new_records))
        for i, r in enumerate(rows):
            is_same_address = builtins.id(r) == builtins.id(new_records[i])
            test_ok = (not is_same_address)
            test_desc = 'Test for NOT SAME.'
            assert test_ok, \
                test_desc + ' Record #' + str(i) + ' same address = ' + str(is_same_address) + ': ' \
                + str(builtins.id(r)) + ' and ' + str(builtins.id(new_records[i])) + ' for record ' + str(r)

        for i, (cond, query, len_exp_recs, exp_remaining_ids, exp_remaining_texts) in enumerate([
            ({'and': True, 'exact': True}, {'id': 101}, 1, [100, 102, 103, 103, 104], None),
            ({'and': True, 'exact': True}, {'text': 'sentence B'}, 2, [100, 103, 104],  ['sentence A', 'sentENce B', 'this is sentence B']),
            ({'and': False, 'exact': True}, {'id': 104, 'text': 'sentence A'}, 2, [103], ['sentENce B',]),
            ({'and': False, 'exact': False}, {'text': 'B'}, 1, [], []),
        ]):
            self.logger.info('Records remaining now ' + str(ds.get_all()))
            recs_to_be_deleted = ds.get(
                match_phrase = query,
                match_condition = cond,
            )
            self.logger.info(
                '#' + str(i) + ' Records to be deleted for query "' + str(query) + '": ' + str(recs_to_be_deleted)
            )
            assert len(recs_to_be_deleted) == len_exp_recs, \
                '#' + str(i) + ' For id "' + str(id) + '" got != expected ' + str(len_exp_recs) + ' record, but ' \
                + str(len(recs_to_be_deleted)) + ': ' + str(recs_to_be_deleted)
            ds.delete(
                match_phrase = query,
                match_condition = cond,
            )
            self.logger.info('#' + str(i) + ' After DELETE ' + str(query) + '..')
            rows = ds.get_all()
            remaining_ids = [r['id'] for r in rows]
            remaining_texts = [r['text'] for r in rows]
            [print(r) for r in rows]
            assert remaining_ids == exp_remaining_ids, \
                '#' + str(i) + ' Remaining ids ' + str(remaining_ids) + ' not expected ' + str(exp_remaining_ids)

            if exp_remaining_texts is not None:
                assert remaining_texts == exp_remaining_texts, \
                    '#' + str(i) + ' Remaining texts ' + str(remaining_texts) \
                    + ' not expected ' + str(exp_remaining_texts)

        print('ALL TESTS PASSED OK')


if __name__ == '__main__':
    er = Env()
    Env.set_env_vars_from_file(env_filepath= er.REPO_DIR + '/.env.fitxf.math.ut')
    Logging.get_logger_from_env_var()
    dbp = DbParams.get_db_params_from_envvars(
        identifier = 'demo',
        db_create_tbl_sql = '',
        db_table = 'testtable'
    )
    exit(0)



import logging
import torch
import math
from fitxf.math.utils.Logging import Logging


class ClassifierArcInterface:

    def __init__(
            self,
            # old state
            model_filepath: str = None,
            in_features: int = None,
            out_features: int = None,
            n_hidden_features: int = 100,
            hidden_functions: list = (torch.nn.Linear, torch.nn.Linear, torch.nn.Linear),
            activation_functions: list = (torch.nn.ReLU, torch.nn.ReLU, torch.nn.Softmax),
            loss_function = torch.nn.CrossEntropyLoss,
            dropout_rate: float = 0.2,
            learning_rate: float = 0.0001,
            logger = None,
    ):
        self.model_filepath = model_filepath
        self.in_features = in_features
        self.out_features = out_features
        self.n_hidden_features = n_hidden_features if type(n_hidden_features) in [list, tuple] \
            else [n_hidden_features, int(round(n_hidden_features / 2))]
        self.hidden_functions = hidden_functions
        self.activation_functions = activation_functions
        self.loss_function = loss_function
        self.dropout_rate = dropout_rate
        self.learning_rate = learning_rate
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def from_old_states(
            self,
            model_filepath: str,
    ):
        raise Exception('Must be implemented by child class')

    def save_states(
            self,
            model_filepath: str,
            additional_info: dict,
    ):
        raise Exception('Must be implemented by child class')

    # just an array of batch tuples
    #    e.g. [(X_batch_1, y_batch_1), (X_batch_2, y_batch_2), ...]
    # or if attention masks included
    #    e.g. [(X_batch_1, attn_mask_batch_1, y_batch_1), (X_batch_2, attn_mask_batch_2, y_batch_2), ...]
    def create_dataloader(
            self,
            X: torch.Tensor,
            y: torch.Tensor,
            # if None, means we don't convert to onehot (possibly caller already done that, or not required)
            y_num_classes: int = None,
            batch_size: int = 32,
            eval_percent: float = 0.2,
            include_attn_mask: bool = False,
    ):
        if y_num_classes is not None:
            y_onehot_or_value = torch.nn.functional.one_hot(
                y.to(torch.int64),
                num_classes = y_num_classes,
            ).to(torch.float)
        else:
            y_onehot_or_value = y

        dataloader_train = []
        idx_batch = 0
        while True:
            j = idx_batch + batch_size
            if j-1 > len(X):
                break
            X_batch = X[idx_batch:j]
            y_onehot_batch = y_onehot_or_value[idx_batch:j]
            if include_attn_mask:
                attn_mask = torch.ones(size=X_batch.shape)
                dataloader_train.append((X_batch, attn_mask, y_onehot_batch))
            else:
                dataloader_train.append((X_batch, y_onehot_batch))
            idx_batch = j

        cutoff_batch_idx = math.floor(len(dataloader_train) * (1 - eval_percent))
        dataloader_eval = dataloader_train[cutoff_batch_idx:]
        dataloader_train = dataloader_train[:cutoff_batch_idx]
        self.logger.info(
            'Created data loader (train/eval) of length ' + str(len(dataloader_train))
            + ' / ' + str(len(dataloader_eval)) + ' of batch sizes ' + str(batch_size)
        )

        # record where we cut off train/eval
        trn_eval_cutoff_idx = cutoff_batch_idx * batch_size
        return dataloader_train, dataloader_eval, trn_eval_cutoff_idx

    def to_onehot(
            self,
            y: torch.Tensor,
    ):
        n_values = torch.max(y.to(torch.int)) + 1
        onehot = torch.eye(n_values)[y.to(torch.int)]
        onehot_f = onehot.to(torch.float)
        return onehot_f
        # alternatively, can also use the torch builtin function
        # return torch.nn.functional.one_hot(
        #     y.to(torch.int64),
        #     num_classes = y_num_classes,
        # ).to(torch.float)

    def forward(
            self,
            x: torch.Tensor,
    ):
        raise Exception('Must be implemented by child class')

    def fit(
            self,
            X: torch.Tensor,
            y: torch.Tensor,
            is_categorical: bool = True,
            # if None, means we don't convert to onehot (possibly caller already done that, or not required)
            num_categories: int = None,
            # the smaller the batch size, the smaller the losses will be during training
            batch_size: int = 32,
            epochs: int = 100,
            eval_percent: float = 0.2,
            # important to prevent over-fitting
            regularization_type = "L2",
    ):
        raise Exception('Must be implemented by child class')

    def predict(
            self,
            X: torch.Tensor,
    ):
        raise Exception('Must be implemented by child class')


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    exit(0)



import warnings
import uuid
import os
import re
from venv import logger

from fitxf.utils import Logging, Profiling, Env
#----------------------------------------------------------------------------------
# IC Section
#----------------------------------------------------------------------------------
# Math
from fitxf.math.fit.transform.FitXformInterface import FitXformInterfaceUnitTest
from fitxf.math.fit.transform.FitXformUnitTest import FitXformUnitTest
from fitxf.math.fit.utils.FitUtilsUnitTest import FitUtilsUt
from fitxf.math.fit.arc.ClassifierArcUnitTest import ClassifierArcUnitTest
from fitxf.math.fit.arc.ClassifierArcRf import ClassifierArcRf
from fitxf.math.fit.utils.TensorUtils import TensorUtilsUnitTest
from fitxf.math.fit.utils.MathUtils import MathUtilsUnitTest
from fitxf.math.fit.utils.PatternSearchUnitTest import PatternSearchUnitTest
from fitxf.math.fit.cluster.ClusterUnitTest import ClusterUnitTest
from fitxf.math.fit.cluster.ClusterCosineUT import ClusterCosineUnitTest
from fitxf.math.graph.GraphUtilsUnitTest import GraphUtilsUnitTest
from fitxf.math.utils.PkgVersion import PkgVersion
# Utils
from fitxf.math.utils.Lock import LockUnitTest
from fitxf.math.utils.ObjPers import UnitTestObjectPersistence
from fitxf.math.utils.Singleton import SingletonUnitTest
from fitxf.math.utils.StringVar import StringVarUnitTest
# Algo
from fitxf.math.algo.encoding.Base64 import Base64UnitTest


class RepoUnitTest:

    def __init__(
            self,
    ):
        self.env_repo = Env()

        self.keys_dir = self.env_repo.REPO_DIR + '/keys'
        self.lm_cache_folder = self.env_repo.MODELS_PRETRAINED_DIR
        self.document_folder = self.env_repo.REPO_DIR + '/data/sample_docs'
        self.tmp_dir = os.environ["TEMP_DIR"]
        self.logger = Logging.get_logger_from_env_var()

        pkg_utils = PkgVersion(logger=self.logger)
        verdict, (self.py_maj_ver, self.py_min_ver)= pkg_utils.check_python(version="3.12", return_version=True)

        rand_str_uuid = str(uuid.uuid4())
        if (self.py_maj_ver == 3) and (self.py_min_ver <= 11):
            rand_str = re.sub(pattern=".*[\-]", repl="", string=rand_str_uuid)
        else:
            rand_str = re.sub(pattern=".*[-]", repl="", string=rand_str_uuid)
        self.db_test_table_or_index = 'nwae-math.repo-unit-test.' + str(rand_str)

        warnings.filterwarnings("ignore", message="Unverified HTTPS request")
        return

    def test(self):
        profiler = Profiling(logger=self.logger)
        test_record = {}

        t00 = profiler.start()

        for cls in [
            # Math
            FitXformInterfaceUnitTest, FitXformUnitTest, FitUtilsUt, TensorUtilsUnitTest, # HomomorphismUnitTest,
            MathUtilsUnitTest, PatternSearchUnitTest, ClusterUnitTest, ClusterCosineUnitTest,
            GraphUtilsUnitTest,
            ClassifierArcUnitTest,
            # Utils
            LockUnitTest, SingletonUnitTest, StringVarUnitTest,
            UnitTestObjectPersistence,
            # Datastore
            # This test can only run if you already set up Elasticsearch locally
            # MemoryCacheUnitTest, MySqlUnitTest, VecDbUnitTest, VecDbCcrcyTest, VecDbSingletonUnitTest,
            # Algo
            Base64UnitTest, # SortColumnsUnitTest, SortRangesAndOverlapsUnitTest,
            # Models
            # MultiTreeUnitTest,
            # Language
            # LangCharUnitTest, LangUnitTest,
            # Text
            # TextDiffUnitTest,
            # MaskTextSortedUnitTest, AnonymizerUnitTest, RegexPpUnitTest, TxtPreprocessorUnitTest,
            # Language Models
            # LangModelInterfaceTest, LangModelPtUnitTest,
            # Intent
            # ClassifyWrapperUnitTest,
            # InfoRetrievalUnitTest,
        ]:
            t0 = profiler.start()
            # if cls not in [VecDbCcrcyTest]:
            #     continue

            self.logger.info('BEGIN TESTING ' + str(cls))
            if cls == FitUtilsUt:
                ut = FitUtilsUt(logger=self.logger)
                ut.test_map_to_connbr()
                # TODO Uncomment these when we use them
                # ut.test_nn(epochs=5000, plot_graph=False, tol_dif=0.1)
                # ut.test_dist()
            elif cls == FitXformUnitTest:
                ut = FitXformUnitTest(
                    lm_cache_folder = self.lm_cache_folder,
                    logger = self.logger,
                )
                ut.test()
            elif cls == TensorUtilsUnitTest:
                ut = TensorUtilsUnitTest()
                ut.test_norm()
                ut.test_similarity_cosine_and_similarity_distance()
            elif cls == ClassifierArcUnitTest:
                ut = ClassifierArcUnitTest(child_class=ClassifierArcRf, logger=self.logger)
                ut.test()
            else:
                cls(logger=self.logger).test()

            t1 = profiler.stop()
            secs_taken = profiler.get_time_dif_secs(start=t0, stop=t1, decimals=10)
            time_taken = str(round(secs_taken, 2))+'s' if secs_taken > 0.01 else \
                    str(round(secs_taken*1000000, 2))+'μs'
            test_record[cls] = {'secs_taken': time_taken}

        self.logger.info('------------------------------------------------------------------------------------')
        self.logger.info('OK DONE ' + str(len(test_record)) + ' TESTS SUCCESSFULLY')
        [
            self.logger.info(
                '   ' + str(i) + '. ' + str(k) + '\n      --> ' + str(v)
            )
            for i, (k, v) in enumerate(test_record.items())
        ]
        self.logger.info('Total secs taken ' + str(profiler.get_time_dif_secs(start=t00, stop=profiler.stop(), decimals=2)))


if __name__ == '__main__':
    env_repo = Env()
    Env.set_env_vars_from_file(env_filepath=env_repo.REPO_DIR + '/.env.fitxf.math.ut')

    RepoUnitTest().test()
    exit(0)



import logging
import torch
import os
import re
import traceback
import requests
import numpy as np
from io import BytesIO
from fitxf import TensorUtils
from PIL import Image
from fitxf.math.lang.encode.LangModelInterface import LangModelInterface as LmInterface
from transformers import AutoImageProcessor, AutoModel
from fitxf.math.utils.Env import Env
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Lock import Lock


class ImgPt(LmInterface):

    RETURN_TENSORS = 'np'

    DEFAULT_MODEL_NAME = 'google/vit-base-patch16-224'
    DEFAULT_CHANNEL = 'RGB'

    def __init__(
            self,
            model_name = None,
            cache_folder = None,
            include_tokenizer = False,
            logger = None,
    ):
        super().__init__(
            model_name = model_name,
            cache_folder = cache_folder,
            include_tokenizer = include_tokenizer,
            logger = logger,
        )

        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.tensor_utils = TensorUtils(logger=self.logger)
        self.use_serverless = False

        self.__mutex_name_model = 'modelpt'
        self.__lock_mutexes = Lock(mutex_names=[self.__mutex_name_model], logger=self.logger)

        if self.model_name is None:
            self.model_name = self.DEFAULT_MODEL_NAME

        # User may pass in model downloaded path
        if os.path.isdir(str(self.model_name)):
            self.model_path = self.model_name
        else:
            self.model_path = self.cache_folder + '/' + self.model_name

        assert os.path.isdir(self.model_path), 'Not a directory "' + str(self.model_path) + '"'
        self.logger.info('Model name "' + str(self.model_name) + '" path "' + str(self.model_path) + '"')

        self.logger.info(
            'Lang model "' + str(self.model_name) + '" with cache folder "' + str(self.cache_folder)
            + '", name_or_path "' + str(self.model_path) + '", device "' + str(self.device) + '"'
        )
        self.processor = AutoImageProcessor.from_pretrained(
            pretrained_model_name_or_path = self.model_path,
        )
        self.logger.info(
            'OK processor for model "' + str(self.model_path) + '", cache folder "' + str(self.cache_folder) + '"'
        )
        self.model = AutoModel.from_pretrained(
            pretrained_model_name_or_path = self.model_path,
        ).to(self.device)
        self.logger.info(
            'OK Model "' + str(self.model_path) + '", cache folder "' + str(self.cache_folder) + '"'
        )
        return

    def download_image(
            self,
            url,
    ):
        try:
            response = requests.get(url)
            img = Image.open(BytesIO(response.content)).convert(self.DEFAULT_CHANNEL)
            # img.save('img_' + str(i) + '.bmp')
            np_image = np.array(img)
            # TODO resize
            return np_image
        except Exception as ex:
            errmsg = 'Failed to get image from URL "' + str(url) + '": ' + str(ex) \
                     + ' Stack trace: ' + str(traceback.format_exc())
            self.logger.error(errmsg)
            raise Exception(errmsg)

    def encode(
            self,
            content_list,
            # max length has got no meaning
            maxlen = None,
            return_tensors = 'pt',
            # does not apply here since we can't see the tokenization
            return_attnmasks = False,
            params_other = None,
    ):
        embeddings = []
        for img_xxx in content_list:
            if type(img_xxx) is str:
                if re.match(pattern="^http", string=img_xxx, flags=re.IGNORECASE):
                    img_file_type = 'url'
                else:
                    img_file_type = 'file'

                if img_file_type == 'url':
                    img_data = self.download_image(url=img_xxx)
                else:
                    img = Image.open(img_xxx)
                    img_data = img

                # TODO Need to reshape?
                # img.thumbnail(size=self.default_shape)
                # img_data = img.resize(size=self.default_shape)
                img_data = np.array(img_data)

                self.logger.info(
                    'Converted non array image data "' + str(img_xxx) + '" type "' + str(img_file_type)
                    + '" to type "' + str(type(img_data)) + '"' # numpy array shape ' + str(img_data.shape)
                )
            elif type(img_data) in [torch.Tensor, np.ndarray]:
                img_data = img_xxx
            else:
                raise Exception('Unsupported image type "' + str(type(img_data)) + '"')
            emb = self.encode_image(
                image = img_data,
                return_tensors = return_tensors,
            )
            self.logger.info('Encoded image to shape ' + str(emb.shape))
            embeddings.append(emb)
        if return_tensors == 'pt':
            return (torch.vstack(embeddings), None) if return_attnmasks else torch.vstack(embeddings)
        else:
            return (np.vstack(embeddings), None) if return_attnmasks else np.vstack(embeddings)

    def encode_image(
            self,
            image: np.ndarray,
            # permitted values 'last_hidden_state_mean', 'last_hidden_state_cls',  'pooler_output'
            embedding_method: str = 'last_hidden_state_mean',
            return_tensors = 'pt',
    ):
        assert embedding_method in ['last_hidden_state_mean', 'last_hidden_state_cls',  'pooler_output']
        self.logger.info(
            'Type of image to encode is "' + str(type(image)) + '", return tensors "' + str(return_tensors)
            # + '", shape ' + str(image.shape)
        )
        inputs = self.processor(
            image,
            return_tensors = 'pt',
        ).to(self.device)
        outputs = self.model(**inputs)
        last_hidden_state = outputs.last_hidden_state
        if embedding_method in ['last_hidden_state_mean']:
            torch_embedding = last_hidden_state.mean(dim=1)
        elif embedding_method in ['last_hidden_state_cls']:
            torch_embedding = last_hidden_state[:, 0, :]
        else:
            self.logger.warning(
                'This method produces varying values for the same inputs!! This is due to uninitialized pooler weights.'
            )
            # An alternative representation is indeed the pooler_output,
            # which takes the embedding of the first special CLS token from the last_hidden_state,
            # and applies a layernorm to it as seen here
            # https://github.com/huggingface/transformers/blob/95754b47a6d4fbdad3440a45762531e8c471c528/src/transformers/models/clip/modeling_clip.py#L865
            torch_embedding = outputs.pooler_output

        if return_tensors == 'pt':
            return torch_embedding
        else:
            return torch_embedding.detach().numpy()


if __name__ == '__main__':
    er = Env()
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/.env.fitxf.math.ut')
    urls = [
        'http://images.cocodataset.org/val2017/000000039769.jpg',
        'https://img.freepik.com/free-photo/large-set-isolated-vegetables-white-background_485709-44.jpg?t=st=1729854477~exp=1729858077~hmac=55e5cf830d6a663aac9875a3de6ee54bf4a38ab8cec3400de6732bd94ea2d444&w=740',
        'https://upload.wikimedia.org/wikipedia/commons/b/b3/Blackpink_Ros%C3%A9_Rimowa_1.jpg',
    ]
    # image = Image.open(requests.get(urls[0], stream=True).raw)
    # np_image = np.array(image)
    # print(image, np_image)
    # print('Image type "' + str(type(image)) + '", shape ' + str(np_image.shape))

    pt = ImgPt(
        cache_folder = er.MODELS_PRETRAINED_DIR,
        logger = Logging.get_default_logger(log_level=logging.INFO, propagate=False),
    )
    embed_repeats = []

    # check that embedding remains the same value for the same inputs
    for i in range(2):
        embed = pt.encode(
            text_list = urls,
            return_tensors = 'np',
            return_attnmasks = False,
        )
        embed_repeats.append(embed)
        print(i, embed[:,-5:])
        print(i, type(embed), embed.shape)
        if i > 0:
            assert np.sum(embed**2 - embed_repeats[-1]**2) < 0.0000000001
    exit(0)



import logging
from fitxf.math.utils.Lock import Logging
from fitxf.math.utils.Env import Env


class DbLang:

    def __init__(
            self,
            logger: Logging = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    @staticmethod
    def get_db_syntax_create_table_mysql(
            tablename: str,
            # ["`name` VARCHAR(255) NOT NULL", ...]
            columns: list,
    ):
        assert len(columns) > 0
        sql = "CREATE TABLE `" + str(tablename) + "` ("
        edge = len(columns) - 1
        for i, col_str in enumerate(columns):
            sql += '\n   ' + str(col_str)
            sql += "," if i != edge else ""
        sql += '\n)'
        return sql

    def get_db_syntax_create_table(
            self,
            db_type: str,
            tablename: str,
            columns: list,
    ):
        if db_type == 'mysql':
            return self.get_db_syntax_create_table_mysql(tablename=tablename, columns=columns)
        else:
            return None


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    er = Env(logger=lgr)
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/.env.fitxf.math.ut.mysql')

    s = DbLang(logger=lgr).get_db_syntax_create_table(
        db_type = 'mysql',
        tablename = "<TABLENAME>",
        columns = [
            "`<CONTENT>` TEXT NOT NULL",
            "`<LABEL_USER>` varchar(255) NOT NULL",
            "`<LABEL_NUMBER>` int NOT NULL",
            "`<ENCODING_B64>` TEXT NOT NULL",
        ],
    )
    print(s)
    exit(0)
